import logging as logger
from pathlib import Path
import os

log_dir = (Path(__file__).parent.parent / 'logs').resolve()
log_file = (log_dir / 'hub.log').resolve()
try:
  if not os.path.isfile(log_file):
    os.mkdir(log_dir)
    with open(log_file, 'x'):
      pass
  logger.basicConfig(filename=log_file, format="%(asctime)s:%(levelname)s:{%(pathname)s:%(funcName)s:%(lineno)d} "
                                               "- %(message)s", level=logger.DEBUG)
except IOError as err:
  print(f'I/O exception: {err}')



"""
City repository with database CRUD operations
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Peter Yefi peteryefi@gmail.com
"""

import datetime
import pickle
from typing import Union, Dict

from sqlalchemy import select
from sqlalchemy.exc import SQLAlchemyError

from hub.city_model_structure.city import City as CityHub
from hub.hub_logger import logger
from hub.persistence import Repository
from hub.persistence.models import City as Model
from hub.persistence.models import CityObject
from hub.version import __version__


class City(Repository):
  _instance = None

  def __init__(self, db_name: str, dotenv_path: str, app_env: str):
    super().__init__(db_name, dotenv_path, app_env)

  def __new__(cls, db_name, dotenv_path, app_env):
    """
    Implemented for a singleton pattern
    """
    if cls._instance is None:
      cls._instance = super(City, cls).__new__(cls)
    return cls._instance

  def insert(self, city: CityHub, application_id, user_id: int) -> Union[Model, Dict]:
    """
    Inserts a city
    :param city: The complete city instance
    :param application_id: Application id owning the instance
    :param user_id: User id owning the instance
    :return: City and Dictionary
    """
    try:
      db_city = Model(
        pickle.dumps(city),
        city.name,
        city.level_of_detail.geometry,
        'None' if city.climate_file is None else city.climate_file,
        application_id,
        user_id,
        __version__)

      self.session.add(db_city)
      self.session.flush()
      for building in city.buildings:
        object_usage = ''
        for internal_zone in building.internal_zones:
          for usage in internal_zone.usages:
            object_usage = f'{object_usage}{usage.name}_{usage.percentage} '
        object_usage = object_usage.rstrip()
        db_city_object = CityObject(db_city.id,
                                    building.name,
                                    building.alias,
                                    building.type,
                                    building.year_of_construction,
                                    building.function,
                                    object_usage,
                                    building.volume,
                                    building.floor_area)
        self.session.add(db_city_object)
        self.session.flush()
      self.session.commit()
      return db_city
    except SQLAlchemyError as err:
      print(f'An error occurred while creating city: {err}')
      logger.error(f'An error occurred while creating city: {err}')

  def update(self, city_id: int, city: CityHub):
    """
    Updates a city name (other updates makes no sense)
    :param city_id: the id of the city to be updated
    :param city: the city object
    :return:
    """
    try:
      now = datetime.datetime.utcnow()
      self.session.query(Model).filter(Model.id == city_id).update({'name': city.name,'updated': now})
      self.session.commit()
    except SQLAlchemyError as err:
      logger.error(f'Error while updating city: {err}')

  def delete(self, city_id: int):
    """
    Deletes a City with the id
    :param city_id: the city id
    :return: a city
    """
    try:
      self.session.query(Model).filter(Model.id == city_id).delete()
      self.session.commit()
    except SQLAlchemyError as err:
      logger.error(f'Error while fetching city: {err}')

  def get_by_id(self, city_id: int) -> Model:
    """
    Fetch a City based on the id
    :param city_id: the city id
    :return: a city
    """
    try:
      return self.session.execute(select(Model).where(Model.id == city_id)).first()[0]
    except SQLAlchemyError as err:
      logger.error(f'Error while fetching city: {err}')

  def _get_by_hub_version_and_name(self, hub_release: str, city_name: str) -> Model:
    """
    Fetch a City based on the name and hub project
    :param hub_release: the hub release
    :param city_name: the name of the city
    :return: a city
    """
    try:
      return self.session.execute(select(Model)
                                  .where(Model.hub_release == hub_release, Model.name == city_name)
                                  ).first()
    except SQLAlchemyError as err:
      logger.error(f'Error while fetching city: {err}')

  def get_by_name(self, city_name: str) -> [Model]:
    """
    Fetch city based on the name
    :param city_name: the name of the building
    :return: [ModelCity] with the provided name
    """
    try:
      result_set = self.session.execute(select(Model).where(Model.name == city_name))
      return [building[0] for building in result_set]
    except SQLAlchemyError as err:
      logger.error(f'Error while fetching city by name: {err}')

  def get_by_user(self, user_id: int) -> [Model]:
      """
      Fetch city based on the user who created it
      :param user_id: the id of the user
      :return: [ModelCity] with the provided name
      """
      try:
        result_set = self.session.execute(select(Model).where(Model.user_id == user_id))
        return [building[0] for building in result_set]
      except SQLAlchemyError as err:
        logger.error(f'Error while fetching city by name: {err}')



"""
Rhino module parses rhino files and import the geometry into the city model structure
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Guille Gutierrez guillermo.gutierrezmorote@concordia.capip
"""
import numpy as np
from rhino3dm import *
from rhino3dm._rhino3dm import Extrusion
from rhino3dm._rhino3dm import MeshType

from hub.city_model_structure.attributes.point import Point
from hub.city_model_structure.attributes.polygon import Polygon
from hub.city_model_structure.building import Building
from hub.city_model_structure.building_demand.surface import Surface as HubSurface
from hub.city_model_structure.city import City
from hub.helpers.configuration_helper import ConfigurationHelper
from hub.imports.geometry.helpers.geometry_helper import GeometryHelper


class Rhino:
  def __init__(self, path):
    self._model = File3dm.Read(str(path))
    max_float = float(ConfigurationHelper().max_coordinate)
    min_float = float(ConfigurationHelper().min_coordinate)
    self._min_x = self._min_y = self._min_z = max_float
    self._max_x = self._max_y = self._max_z = min_float

  @staticmethod
  def _in_perimeter(wall, corner):
    res = wall.contains_point(Point(corner))
    return res

  @staticmethod
  def _add_hole(solid_polygon, hole):
    first = solid_polygon.points[0]
    points = first + hole.points + solid_polygon.points
    return Polygon(points)

  @staticmethod
  def _solid_points(coordinates) -> np.ndarray:
    solid_points = np.fromstring(coordinates, dtype=float, sep=' ')
    solid_points = GeometryHelper.to_points_matrix(solid_points)

    result = []
    found = False
    for row in solid_points:
      for row2 in result:
        if row[0] == row2[0] and row[1] == row2[1] and row[2] == row2[2]:
          found = True
      if not found:
        result.append(row)
    return solid_points

  def _corners(self, point):
    if point.X < self._min_x:
      self._min_x = point.X
    if point.Y < self._min_y:
      self._min_y = point.Y
    if point.Z < self._min_z:
      self._min_z = point.Z
    if point.X > self._max_x:
      self._max_x = point.X
    if point.Y > self._max_y:
      self._max_y = point.Y
    if point.Z > self._max_z:
      self._max_z = point.Z

  def _add_face(self, face):
    hub_surfaces = []
    _mesh = face.GetMesh(MeshType.Default)
    for i in range(0, len(_mesh.Faces)):
      mesh_faces = _mesh.Faces[i]
      _points = ''
      faces = []
      for index in mesh_faces:
        if index in faces:
          continue
        faces.append(index)
        self._corners(_mesh.Vertices[index])
        _points = _points + f'{_mesh.Vertices[index].X} {_mesh.Vertices[index].Y} {_mesh.Vertices[index].Z} '
      polygon_points = Rhino._solid_points(_points.strip())
      hub_surfaces.append(HubSurface(Polygon(polygon_points), Polygon(polygon_points)))
    return hub_surfaces

  @property
  def city(self) -> City:
    buildings = []
    city_objects = []  # building and "windows"
    windows = []
    _prev_name = ''

    for obj in self._model.Objects:
      name = obj.Attributes.Id
      hub_surfaces = []
      if isinstance(obj.Geometry, Extrusion):
        surface = obj.Geometry
        hub_surfaces = hub_surfaces + self._add_face(surface)
      else:
        for face in obj.Geometry.Faces:
          if face is None:
            break
          hub_surfaces = hub_surfaces + self._add_face(face)
      building = Building(name, hub_surfaces, 'unknown', 'unknown', [])
      city_objects.append(building)
    lower_corner = (self._min_x, self._min_y, self._min_z)
    upper_corner = (self._max_x, self._max_y, self._max_z)
    city = City(lower_corner, upper_corner, 'EPSG:26918')
    for building in city_objects:
      if len(building.surfaces) <= 2:
        # is not a building but a window!
        for surface in building.surfaces:
          # add to windows the "hole" with the normal inverted
          windows.append(Polygon(surface.perimeter_polygon.inverse))
      else:
        buildings.append(building)

    # todo: this method will be pretty inefficient
    for hole in windows:
      corner = hole.coordinates[0]
      for building in buildings:
        for surface in building.surfaces:
          plane = surface.perimeter_polygon.plane
          # todo: this is a hack for dompark project it should not be done this way windows should be correctly modeled
          # if the distance between the wall plane and the window is less than 2m
          # and the window Z coordinate it's between the wall Z, it's a window of that wall
          if plane.distance(corner) <= 2:
            # check if the window is in the right high.
            if surface.upper_corner[2] >= corner[2] >= surface.lower_corner[2]:
              if surface.holes_polygons is None:
                surface.holes_polygons = []
              surface.holes_polygons.append(hole)

    for building in buildings:
      city.add_city_object(building)
    city.level_of_detail.geometry = 3
    return city



"""
ExportsFactory export a city into several formats
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Guille Gutierrez guillermo.gutierrezmorote@concordia.ca
"""

from pathlib import Path
from hub.exports.formats.obj import Obj
from hub.exports.formats.simplified_radiosity_algorithm import SimplifiedRadiosityAlgorithm
from hub.exports.formats.stl import Stl
from hub.hub_logger import logger
from hub.helpers.utils import validate_import_export_type


class ExportsFactory:
  """
  Exports factory class
  """
  def __init__(self, export_type, city, path, target_buildings=None, adjacent_buildings=None):
    self._city = city
    self._export_type = '_' + export_type.lower()
    class_funcs = validate_import_export_type(ExportsFactory)
    if self._export_type not in class_funcs:
      err_msg = f"Wrong export type [{self._export_type}]. Valid functions include {class_funcs}"
      logger.error(err_msg)
      raise Exception(err_msg)
    if isinstance(path, str):
      path = Path(path)
    self._path = path
    self._target_buildings = target_buildings
    self._adjacent_buildings = adjacent_buildings

  @property
  def _citygml(self):
    """
    Export to citygml
    :return: None
    """
    raise NotImplementedError

  @property
  def _collada(self):
    raise NotImplementedError

  @property
  def _stl(self):
    """
    Export the city geometry to stl
    :return: None
    """
    return Stl(self._city, self._path)

  @property
  def _obj(self):
    """
    Export the city geometry to obj
    :return: None
    """
    return Obj(self._city, self._path)

  @property
  def _grounded_obj(self):
    """
    Export the city geometry to obj with grounded coordinates
    :return: None
    """
    return Obj(self._city, self._path).to_ground_points()

  @property
  def _sra(self):
    """
    Export the city to Simplified Radiosity Algorithm xml format
    :return: None
    """
    return SimplifiedRadiosityAlgorithm(self._city, (self._path / f'{self._city.name}_sra.xml'),
                                        target_buildings=self._target_buildings)

  def export(self):
    """
    Export the city given to the class using the given export type handler
    :return: None
    """
    return getattr(self, self._export_type, lambda: None)

  def export_debug(self):
    """
    Export the city given to the class using the given export type handler
    :return: None
    """
    return Obj(self._city, self._path)



"""
WaterToWaterHPParameters import the heat pump information
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Peter Yefi peteryefi@gmail.com
"""

import pandas as pd
from typing import Dict
from hub.city_model_structure.energy_systems.water_to_water_hp import WaterToWaterHP
from hub.city_model_structure.energy_system import EnergySystem
from scipy.optimize import curve_fit
import numpy as np
from typing import List


class WaterToWaterHPParameters:
  """
  WaterToWaterHPParameters class
  """

  def __init__(self, city, base_path):
    self._city = city
    self._base_path = (base_path / 'heat_pumps/water_to_water.xlsx')

  def _read_file(self) -> Dict:
    """
    reads xlsx file containing water to water heat pump information
    into a dictionary
    :return : Dict
    """
    xl_file = pd.ExcelFile(self._base_path)
    heat_pump_dfs = {sheet_name: xl_file.parse(sheet_name)
                     for sheet_name in xl_file.sheet_names}

    hp_data = {}
    flow_rates = {
      '156': [2.84, 4.23, 5.68],
      '256': [4.73, 7.13, 9.446],
      '335': [6.62, 9.97, 12.93],
    }

    for sheet, dataframe in heat_pump_dfs.items():

      df = heat_pump_dfs[sheet].dropna(axis=1, how='all')
      df = df.iloc[3:, 6:35]

      if '156' in sheet:
        hp_data[sheet] = self._extract_required_hp_data(df, [0, 10, 25, 40, 55, 67], flow_rates['156'])
      elif '256' in sheet:
        hp_data[sheet] = self._extract_required_hp_data(df, [0, 9, 24, 39, 54, 66], flow_rates['256'])
      elif '335' in sheet:
        hp_data[sheet] = self._extract_required_hp_data(df, [0, 11, 26, 41, 56, 69], flow_rates['335'])

    return hp_data

  def _extract_required_hp_data(self, dataframe, ranges, flow_rates):
    """
    Extracts 156 Kw water to water heat pump data
    :param dataframe: dataframe containing all data
    :param ranges: the range of values to extract
    :param flow_rates: the flow rates of water through pump
    :return: Dict
    """
    # extract data rows and columns
    data = {'tc': self._extract_hp_data(dataframe, [1, 11, 21], ranges),
            'pd': self._extract_hp_data(dataframe, [2, 12, 22], ranges),
            'lwt': self._extract_hp_data(dataframe, [5, 15, 25], ranges),
            'fr': (self._extract_flow_and_ewt(dataframe, ranges, [1, 11, 21], flow_rates))[0],
            'ewt': (self._extract_flow_and_ewt(dataframe, ranges, [1, 11, 21], flow_rates))[1]}
    # range values for extracting data
    return data

  @staticmethod
  def _extract_hp_data(df, columns, ranges):
    """
    Extract variable specific (LWT, PD or TC) data from water to water hp
    :param df: the dataframe
    :param columns: the columns to extract data from
    :param ranges: the range of values to extract
    :return: List
    """
    data = pd.concat([df.iloc[ranges[0]:ranges[1], columns[0]], df.iloc[ranges[0]:ranges[1], columns[1]]])
    data = pd.concat([df.iloc[ranges[0]:ranges[1], columns[2]], data])
    for i in range(1, 5):
      data = pd.concat([df.iloc[ranges[i]:ranges[i + 1], columns[0]], data])
      data = pd.concat([df.iloc[ranges[i]:ranges[i + 1], columns[1]], data])
      data = pd.concat([df.iloc[ranges[i]:ranges[i + 1], columns[2]], data])

    return data.dropna().values.tolist()

  @staticmethod
  def _extract_flow_and_ewt(df, ranges, columns, flow_rates):
    """
    Create the flow and ewt data based on the length of the various
    columns for the variables being extracted
    :param df: the dataframe
    :param ranges: the range of values to extract
    :param columns: the columns to extract data from
    :param flow_rates: flow rate values
    :return:
    """
    ewt_values = [-1.111111111, 4.444444444, 10, 15.55555556, 21.11111111]
    length = [len(df.iloc[ranges[0]:ranges[1], columns[0]].dropna()),
              len(df.iloc[ranges[0]:ranges[1], columns[1]].dropna()),
              len(df.iloc[ranges[0]:ranges[1], columns[2]].dropna())]

    ewt_data = np.repeat(ewt_values[0], sum(length))
    flow_rates_data = np.repeat(flow_rates, length)

    for i in range(1, 5):
      length = [len(df.iloc[ranges[i]:ranges[i + 1], columns[0]].dropna()),
                len(df.iloc[ranges[i]:ranges[i + 1], columns[1]].dropna()),
                len(df.iloc[ranges[i]:ranges[i + 1], columns[2]].dropna())]
      flow_rates_data = np.append(flow_rates_data, np.repeat(flow_rates, length))
      ewt_data = np.append(ewt_data, np.repeat(ewt_values[i], sum(length)))

    return flow_rates_data.tolist(), ewt_data.tolist()

  def enrich_city(self):
    """
    Enriches the city with information from file
    """
    heap_pump_data = self._read_file()
    for model, data in heap_pump_data.items():
      heat_pump = WaterToWaterHP()
      heat_pump.model = model.strip()
      heat_pump.total_cooling_capacity = data['tc']
      heat_pump.power_demand = data['pd']
      heat_pump.flow_rate = data['fr']
      heat_pump.entering_water_temp = data['ewt']
      heat_pump.leaving_water_temp = data['lwt']
      heat_pump.power_demand_coff = self._compute_coefficients(data)
      energy_system = EnergySystem(heat_pump.model, [])
      energy_system.water_to_water_hp = heat_pump
      self._city.add_city_object(energy_system)
    return self._city

  def _compute_coefficients(self, heat_pump_data: Dict) -> List[float]:
    """
    Compute heat output and electrical demand coefficients
    from heating performance data
    :param heat_pump_data: a dictionary of heat pump data.
    :return: Tuple[Dict, Dict]
    """
    demand = [i / j for i, j in zip(heat_pump_data['tc'], heat_pump_data['pd'])]

    # Compute heat output coefficients
    popt, _ = curve_fit(self._objective_function, [heat_pump_data['ewt'], heat_pump_data['lwt'], heat_pump_data['fr']],
                        demand)
    return popt.tolist()

  @staticmethod
  def _objective_function(xdata: List, a1: float, a2: float, a3: float, a4: float, a5: float, a6: float,
                          a7: float, a8: float, a9: float, a10: float, a11: float) -> float:
    """
    Objective function for computing coefficients
    :param xdata:
    :param a1: float
    :param a2: float
    :param a3: float
    :param a4: float
    :param a5: float
    :param a6: float
    :param a7: float
    :param a8: float
    :param a9: float
    :param a10: float
    :param a11: float
    :return:
    """
    x, y, t = xdata
    return (a1 * x ** 2) + (a2 * x) + (a3 * y ** 2) + (a4 * y) + (a5 * t ** 2) + (a6 * t) + (a7 * x * y) + (
          a8 * x * t) + (a9 * y * t) + (a10 * x * y * t) + a11



"""
Bus depot module
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""

from hub.city_model_structure.transport.bus_node import BusNode


class BusDepot(BusNode):
  """
  BusDepot class
  """

  def __init__(self, name, coordinates, edges=None):
    super().__init__(name, coordinates, edges=edges, node_type='BusDepot')
    self._number_of_charging_poles = None
    self._number_of_available_buses = None

  @property
  def number_of_charging_poles(self):
    """
    Add explanation here
    :return: add type of variable here
    """
    return self._number_of_charging_poles

  @property
  def number_of_available_buses(self):
    """
    Add explanation here
    :return: add type of variable here
    """
    return self._number_of_available_buses



"""
Sensor measure module
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Guille Gutierrez guillermo.gutierrezmorote@concordia.ca
"""

class SensorMeasure:
  def __init__(self, latitude, longitude, utc_timestamp, value):
    self._latitude = latitude
    self._longitude = longitude
    self._utc_timestamp = utc_timestamp
    self._value = value

  @property
  def latitude(self):
    """
    Get measure latitude
    """
    return self._latitude

  @property
  def longitude(self):
    """
    Get measure longitude
    """
    return self._longitude

  @property
  def utc_timestamp(self):
    """
    Get measure timestamp in utc
    """
    return self._utc_timestamp

  @property
  def value(self):
    """
    Get sensor measure value
    """
    return self._value



"""
Building test
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Atiya atiya.atiya@mail.concordia.ca
"""
from pathlib import Path
from unittest import TestCase
from hub.imports.geometry_factory import GeometryFactory
from hub.imports.life_cycle_assessment_factory import LifeCycleAssessment


class TestLifeCycleAssessment(TestCase):
  """
  TestBuilding TestCase 1
  """
  def setUp(self) -> None:
    """
    Test setup
    :return: None
    """
    self._city_gml = None
    self._example_path = (Path(__file__).parent / 'tests_data').resolve()

  def test_fuel(self):
    city_file = "tests_data/C40_Final.gml"
    city = GeometryFactory('citygml', path=city_file).city
    LifeCycleAssessment('fuel', city).enrich()
    for fuel in city.fuels:
      self.assertTrue(len(city.fuels) > 0)

  def test_vehicle(self):
    city_file = "tests_data/C40_Final.gml"
    city = GeometryFactory('citygml', path=city_file).city
    LifeCycleAssessment('vehicle', city).enrich()
    for vehicle in city.vehicles:
      self.assertTrue(len(city.vehicles) > 0)

  def test_machine(self):
    city_file = "tests_data/C40_Final.gml"
    city = GeometryFactory('citygml', path=city_file).city
    LifeCycleAssessment('machine', city).enrich()
    for machine in city.machines:
      self.assertTrue(len(city.machines) > 0)

  def test_material(self):
    city_file = "tests_data/C40_Final.gml"
    city = GeometryFactory('citygml', path=city_file).city
    LifeCycleAssessment('material', city).enrich()
    for material in city.lca_materials:
      self.assertTrue(len(city.lca_materials) > 0)







"""
Simulation results repository with database CRUD operations
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Guille Gutierrez Guillermo.GutierrezMorote@concordia.ca
"""

import datetime
from typing import Union, Dict

from sqlalchemy import select
from sqlalchemy.exc import SQLAlchemyError

from hub.hub_logger import logger
from hub.persistence import Repository
from hub.persistence.models import SimulationResults as Model
from hub.persistence.models import City
from hub.persistence.models import CityObject

from hub.city_model_structure.building import Building


class SimulationResults(Repository):
  _instance = None

  def __init__(self, db_name: str, dotenv_path: str, app_env: str):
    super().__init__(db_name, dotenv_path, app_env)

  def __new__(cls, db_name, dotenv_path, app_env):
    """
    Implemented for a singleton pattern
    """
    if cls._instance is None:
      cls._instance = super(SimulationResults, cls).__new__(cls)
    return cls._instance

  def insert(self, name: str, values: str, city_id=None, city_object_id=None) -> Union[Model, Dict]:
    """
    Inserts simulations results linked either with a city as a whole or with a city object
    :param name: results name
    :param values: the simulation results in json format
    :param city_id: optional city id
    :param city_object_id: optional city object id
    :return SimulationResults or Dictionary
    """
    if city_id is not None:
      city = self.get_city(city_id)
      if city is None:
        return {'message': f'City does not exists'}
    else:
      city_object = self.get_city_object(city_object_id)
      if city_object is None:
        return {'message': f'City object does not exists'}
    try:
      simulation_result = Model(name=name,
                                values=values,
                                city_id=city_id,
                                city_object_id=city_object_id)
      self.session.add(simulation_result)
      self.session.flush()
      self.session.commit()
      return simulation_result
    except SQLAlchemyError as err:
      logger.error(f'An error occurred while creating city_object: {err}')

  def update(self, name: str, values: str, city_id=None, city_object_id=None) -> Union[Dict, None]:
    """
    Updates simulation results for a city or a city object
    :param name: The simulation results tool and workflow name
    :param values: the simulation results in json format
    :param city_id: optional city id
    :param city_object_id: optional city object id
    :return: None or dictionary
    """
    try:
      if city_id is not None:
        self.session.query(Model).filter(Model.name == name, Model.city_id == city_id).update(
          {
            'values': values,
            'updated': datetime.datetime.utcnow()
          })
        self.session.commit()
      elif city_object_id is not None:
        self.session.query(Model).filter(Model.name == name, Model.city_object_id == city_object_id).update(
          {
            'values': values,
            'updated': datetime.datetime.utcnow()
          })
        self.session.commit()
      else:
        return {'message': 'Missing either city_id or city_object_id'}
    except SQLAlchemyError as err:
      logger.error(f'Error while updating city object: {err}')
      return {'message': 'Error occurred while updating application'}

  def delete(self, name: str, city_id=None, city_object_id=None):
    """
    Deletes an application with the application_uuid
    :param name: The simulation results tool and workflow name
    :param city_id: The id for the city owning the simulation results
    :param city_object_id: the id for the city_object ownning these simulation results

    :return: None
    """
    try:
      if city_id is not None:
        self.session.query(Model).filter(Model.name == name, Model.city_id == city_id).delete()
        self.session.commit()
      elif city_object_id is not None:
        self.session.query(Model).filter(Model.name == name, Model.city_object_id == city_object_id).delete()
        self.session.commit()
      else:
        return {'message': 'Missing either city_id or city_object_id'}
    except SQLAlchemyError as err:
      logger.error(f'Error while deleting application: {err}')

  def get_city(self, city_id) -> [City]:
    """
    Fetch a city object based city id
    :param city_id: a city identifier
    :return: [City] with the provided city_id
    """
    try:
      return self.session.execute(select(City).where(City.id == city_id)).first()
    except SQLAlchemyError as err:
      logger.error(f'Error while fetching city by city_id: {err}')

  def get_city_object(self, city_object_id) -> [CityObject]:
    """
    Fetch a city object based city id
    :param city_object_id: a city object identifier
    :return: [CityObject] with the provided city_object_id
    """
    try:
      return self.session.execute(select(CityObject).where(CityObject.id == city_object_id)).first()
    except SQLAlchemyError as err:
      logger.error(f'Error while fetching city by city_id: {err}')



"""
Model representation of a city object
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Guille Gutierrez Guillermo.GutierrezMorote@concordia.ca
"""

import datetime

from sqlalchemy import Column, Integer, String, Sequence, ForeignKey, Float
from sqlalchemy import DateTime
from hub.persistence.configuration import Models

class CityObject(Models):
  """
  A model representation of an application
  """
  __tablename__ = 'city_object'
  id = Column(Integer, Sequence('application_id_seq'), primary_key=True)
  city_id = Column(Integer, ForeignKey('city.id'), nullable=False)
  name = Column(String, nullable=False)
  alias = Column(String, nullable=True)
  type = Column(String, nullable=False)
  year_of_construction = Column(Integer, nullable=True)
  function = Column(String, nullable=True)
  usage = Column(String, nullable=True)
  volume = Column(Float, nullable=False)
  area = Column(Float, nullable=False)
  created = Column(DateTime, default=datetime.datetime.utcnow)
  updated = Column(DateTime, default=datetime.datetime.utcnow)

  def __init__(self, city_id, name, alias, object_type, year_of_construction, function, usage, volume, area):
    self.city_id = city_id
    self.name = name
    self.alias = alias
    self.type = object_type
    self.year_of_construction = year_of_construction
    self.function = function
    self.usage = usage
    self.volume = volume
    self.area = area



"""
Greenery catalog data model Plant percentage class
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Guille Gutierrez guillermo.gutierrezmorote@concordia.ca
"""

from hub.catalog_factories.data_models.greenery.plant import Plant as libs_plant


class PlantPercentage(libs_plant):

  def __init__(self, percentage, plant_category, plant):
    super().__init__(plant_category, plant)
    self._percentage = percentage

  @property
  def percentage(self):
    """
    Get plant percentage
    :return: float
    """
    return self._percentage



"""
Geometry helper
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Guille Gutierrez guillermo.gutierrezmorote@concordia.ca
Code contributors: Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""
import math
import numpy as np
import requests
from trimesh import Trimesh
from trimesh import intersections
from hub.city_model_structure.attributes.polygon import Polygon
from hub.city_model_structure.attributes.polyhedron import Polyhedron
from hub.helpers.location import Location
from hub.helpers.configuration_helper import ConfigurationHelper


class GeometryHelper:
  """
  Geometry helper class
  """
  srs_transformations = {
    'urn:adv:crs:ETRS89_UTM32*DE_DHHN92_NH': 'epsg:25832'
  }

  def __init__(self, delta=0, area_delta=0):
    self._delta = delta
    self._area_delta = area_delta

  @staticmethod
  def adjacent_locations(location1, location2):
    """
    Determine when two attributes may be adjacent or not based in the dis
    :param location1:
    :param location2:
    :return: Boolean
    """
    max_distance = ConfigurationHelper().max_location_distance_for_shared_walls
    return GeometryHelper.distance_between_points(location1, location2) < max_distance

  def almost_same_area(self, area_1, area_2):
    """
    Compare two areas and decides if they are almost equal (absolute error under delta)
    :param area_1
    :param area_2
    :return: Boolean
    """
    if area_1 == 0 or area_2 == 0:
      return False
    delta = math.fabs(area_1 - area_2)
    return delta <= self._area_delta

  def is_almost_same_surface(self, surface_1, surface_2):
    """
    Compare two surfaces and decides if they are almost equal (quadratic error under delta)
    :param surface_1: Surface
    :param surface_2: Surface
    :return: Boolean
    """

    # delta is grads an need to be converted into radians
    delta = np.rad2deg(self._delta)
    difference = (surface_1.inclination - surface_2.inclination) % math.pi
    if abs(difference) > delta:
      return False
    # s1 and s2 are at least almost parallel surfaces
    # calculate distance point to plane using all the vertex
    # select surface1 value for the point (X,Y,Z) where two of the values are 0
    minimum_distance = self._delta + 1
    parametric = surface_2.polygon.get_parametric()
    normal_2 = surface_2.normal
    for point in surface_1.points:
      distance = abs(
        (point[0] * parametric[0]) + (point[1] * parametric[1]) + (point[2] * parametric[2]) + parametric[3])
      normal_module = math.sqrt(pow(normal_2[0], 2) + pow(normal_2[1], 2) + pow(normal_2[2], 2))

      if normal_module == 0:
        continue
      distance = distance / normal_module
      if distance < minimum_distance:
        minimum_distance = distance
      if minimum_distance <= self._delta:
        break

    if minimum_distance > self._delta or surface_1.intersect(surface_2) is None:
      return False
    return True

  @staticmethod
  def segment_list_to_trimesh(lines) -> Trimesh:
    """
    Transform a list of segments into a Trimesh
    """
    line_points = [lines[0][0], lines[0][1]]
    lines.remove(lines[0])
    while len(lines) > 1:
      i = 0
      for line in lines:
        i += 1
        if GeometryHelper.distance_between_points(line[0], line_points[len(line_points) - 1]) < 1e-8:
          line_points.append(line[1])
          lines.pop(i - 1)
          break
        if GeometryHelper.distance_between_points(line[1], line_points[len(line_points) - 1]) < 1e-8:
          line_points.append(line[0])
          lines.pop(i - 1)
          break
    polyhedron = Polyhedron(Polygon(line_points).triangulate())
    trimesh = Trimesh(polyhedron.vertices, polyhedron.faces)
    return trimesh

  @staticmethod
  def _merge_meshes(mesh1, mesh2):
    v_1 = mesh1.vertices
    f_1 = mesh1.faces
    v_2 = mesh2.vertices
    f_2 = mesh2.faces
    length = len(v_1)
    v_merge = np.concatenate((v_1, v_2))
    f_merge = np.asarray(f_1)

    for item in f_2:
      point1 = item.item(0) + length
      point2 = item.item(1) + length
      point3 = item.item(2) + length
      surface = np.asarray([point1, point2, point3])
      f_merge = np.concatenate((f_merge, [surface]))

    mesh_merge = Trimesh(vertices=v_merge, faces=f_merge)
    mesh_merge.fix_normals()

    return mesh_merge

  @staticmethod
  def divide_mesh_by_plane(trimesh, normal_plane, point_plane):
    """
    Divide a mesh by a plane
    :param trimesh: Trimesh
    :param normal_plane: [x, y, z]
    :param point_plane: [x, y, z]
    :return: [Trimesh]
    """
    # The first mesh returns the positive side of the plane and the second the negative side.
    # If the plane does not divide the mesh (i.e. it does not touch it or it is coplanar with one or more faces),
    # then it returns only the original mesh.
    # todo: review split method in https://github.com/mikedh/trimesh/issues/235,
    #  once triangulate_polygon in Polygon class is solved

    normal_plane_opp = [None] * len(normal_plane)
    for i in range(0, len(normal_plane)):
      normal_plane_opp[i] = - normal_plane[i]

    section_1 = intersections.slice_mesh_plane(trimesh, normal_plane, point_plane)
    if section_1 is None:
      return [trimesh]
    lines = list(intersections.mesh_plane(trimesh, normal_plane, point_plane))
    cap = GeometryHelper.segment_list_to_trimesh(lines)
    trimesh_1 = GeometryHelper._merge_meshes(section_1, cap)

    section_2 = intersections.slice_mesh_plane(trimesh, normal_plane_opp, point_plane)
    if section_2 is None:
      return [trimesh_1]
    trimesh_2 = GeometryHelper._merge_meshes(section_2, cap)

    return [trimesh_1, trimesh_2]

  @staticmethod
  def get_location(latitude, longitude) -> Location:
    """
    Get Location from latitude and longitude
    """
    url = 'https://nominatim.openstreetmap.org/reverse?lat={latitude}&lon={longitude}&format=json'
    response = requests.get(url.format(latitude=latitude, longitude=longitude))
    if response.status_code != 200:
      # This means something went wrong.
      raise Exception('GET /tasks/ {}'.format(response.status_code))

    response = response.json()
    city = 'Unknown'
    country = 'ca'
    if 'city' in response['address']:
      city = response['address']['city']
    if 'country_code' in response['address']:
      country = response['address']['country_code']
    return Location(country, city)

  @staticmethod
  def distance_between_points(vertex1, vertex2):
    """
    distance between points in an n-D Euclidean space
    :param vertex1: point or vertex
    :param vertex2: point or vertex
    :return: float
    """
    power = 0
    for dimension in range(0, len(vertex1)):
      power += math.pow(vertex2[dimension]-vertex1[dimension], 2)
    distance = math.sqrt(power)
    return distance



"""
Dictionaries module saves all transformations of functions and usages to access the catalogs
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""

import hub.helpers.constants as cte


class Dictionaries:
  """
  Dictionaries class
  """
  _usage_to_hft_usage = {
    cte.RESIDENTIAL: 'residential',
    cte.SINGLE_FAMILY_HOUSE: 'single family house',
    cte.MULTI_FAMILY_HOUSE: 'multifamily house',
    cte.ROW_HOUSE: 'single family house',
    cte.MID_RISE_APARTMENT: 'multifamily house',
    cte.HIGH_RISE_APARTMENT: 'multifamily house',
    cte.OFFICE_AND_ADMINISTRATION: 'office and administration',
    cte.SMALL_OFFICE: 'office and administration',
    cte.MEDIUM_OFFICE: 'office and administration',
    cte.LARGE_OFFICE: 'office and administration',
    cte.COURTHOUSE: 'office and administration',
    cte.FIRE_STATION: 'office and administration',
    cte.PENITENTIARY: 'school with shower',
    cte.POLICE_STATION: 'office and administration',
    cte.POST_OFFICE: 'office and administration',
    cte.LIBRARY: 'office and administration',
    cte.EDUCATION: 'education',
    cte.PRIMARY_SCHOOL: 'school without shower',
    cte.PRIMARY_SCHOOL_WITH_SHOWER: 'school with shower',
    cte.SECONDARY_SCHOOL: 'education',
    cte.UNIVERSITY: 'education',
    cte.LABORATORY_AND_RESEARCH_CENTER: 'laboratory and research centers',
    cte.STAND_ALONE_RETAIL: 'retail',
    cte.HOSPITAL: 'health care',
    cte.OUT_PATIENT_HEALTH_CARE: 'health care',
    cte.HEALTH_CARE: 'health care',
    cte.RETIREMENT_HOME_OR_ORPHANAGE: 'Home for the aged or orphanage',
    cte.COMMERCIAL: 'retail',
    cte.STRIP_MALL: 'retail',
    cte.SUPERMARKET: 'retail shop / refrigerated food',
    cte.RETAIL_SHOP_WITHOUT_REFRIGERATED_FOOD: 'retail',
    cte.RETAIL_SHOP_WITH_REFRIGERATED_FOOD: 'retail shop / refrigerated food',
    cte.RESTAURANT: 'restaurant',
    cte.QUICK_SERVICE_RESTAURANT: 'restaurant',
    cte.FULL_SERVICE_RESTAURANT: 'restaurant',
    cte.HOTEL: 'hotel',
    cte.HOTEL_MEDIUM_CLASS: 'hotel (Medium-class)',
    cte.SMALL_HOTEL: 'hotel',
    cte.LARGE_HOTEL: 'hotel',
    cte.DORMITORY: 'dormitory',
    cte.EVENT_LOCATION: 'event location',
    cte.CONVENTION_CENTER: 'event location',
    cte.HALL: 'hall',
    cte.GREEN_HOUSE: 'green house',
    cte.INDUSTRY: 'industry',
    cte.WORKSHOP: 'industry',
    cte.WAREHOUSE: 'industry',
    cte.WAREHOUSE_REFRIGERATED: 'industry',
    cte.SPORTS_LOCATION: 'sport location',
    cte.SPORTS_ARENA: 'sport location',
    cte.GYMNASIUM: 'sport location',
    cte.MOTION_PICTURE_THEATRE: 'event location',
    cte.MUSEUM: 'event location',
    cte.PERFORMING_ARTS_THEATRE: 'event location',
    cte.TRANSPORTATION: 'n/a',
    cte.AUTOMOTIVE_FACILITY: 'n/a',
    cte.PARKING_GARAGE: 'n/a',
    cte.RELIGIOUS: 'event location',
    cte.NON_HEATED: 'non-heated'
  }

  _usage_to_comnet_usage = {
    cte.RESIDENTIAL: 'BA Multifamily',
    cte.SINGLE_FAMILY_HOUSE: 'BA Multifamily',
    cte.MULTI_FAMILY_HOUSE: 'BA Multifamily',
    cte.ROW_HOUSE: 'BA Multifamily',
    cte.MID_RISE_APARTMENT: 'BA Multifamily',
    cte.HIGH_RISE_APARTMENT: 'BA Multifamily',
    cte.OFFICE_AND_ADMINISTRATION: 'BA Office',
    cte.SMALL_OFFICE: 'BA Office',
    cte.MEDIUM_OFFICE: 'BA Office',
    cte.LARGE_OFFICE: 'BA Office',
    cte.COURTHOUSE: 'BA Courthouse',
    cte.FIRE_STATION: 'BA Fire Station',
    cte.PENITENTIARY: 'BA Penitentiary',
    cte.POLICE_STATION: 'BA Police Station',
    cte.POST_OFFICE: 'BA Post Office',
    cte.LIBRARY: 'BA Library',
    cte.EDUCATION: 'BA School/University',
    cte.PRIMARY_SCHOOL: 'BA School/University',
    cte.PRIMARY_SCHOOL_WITH_SHOWER: 'BA School/University',
    cte.SECONDARY_SCHOOL: 'BA School/University',
    cte.UNIVERSITY: 'BA School/University',
    cte.LABORATORY_AND_RESEARCH_CENTER: 'BA School/University',
    cte.STAND_ALONE_RETAIL: 'BA Retail',
    cte.HOSPITAL: 'BA Hospital',
    cte.OUT_PATIENT_HEALTH_CARE: 'BA Healthcare Clinic',
    cte.HEALTH_CARE: 'BA Healthcare Clinic',
    cte.RETIREMENT_HOME_OR_ORPHANAGE: 'BA Healthcare Clinic',
    cte.COMMERCIAL: 'BA Retail',
    cte.STRIP_MALL: 'BA Retail',
    cte.SUPERMARKET: 'BA Retail',
    cte.RETAIL_SHOP_WITHOUT_REFRIGERATED_FOOD: 'BA Retail',
    cte.RETAIL_SHOP_WITH_REFRIGERATED_FOOD: 'BA Retail',
    cte.RESTAURANT: 'BA Dining: Bar Lounge/Leisure',
    cte.QUICK_SERVICE_RESTAURANT: 'BA Dining: Cafeteria/Fast Food',
    cte.FULL_SERVICE_RESTAURANT: 'BA Dining: Bar Lounge/Leisure',
    cte.HOTEL: 'BA Hotel',
    cte.HOTEL_MEDIUM_CLASS: 'BA Motel',
    cte.SMALL_HOTEL: 'BA Motel',
    cte.LARGE_HOTEL: 'BA Hotel',
    cte.DORMITORY: 'BA Dormitory',
    cte.EVENT_LOCATION: 'BA Convention Center',
    cte.CONVENTION_CENTER: 'BA Convention Center',
    cte.HALL: 'BA Town Hall',
    cte.GREEN_HOUSE: 'n/a',
    cte.INDUSTRY: 'BA Manufacturing Facility',
    cte.WORKSHOP: 'BA Workshop',
    cte.WAREHOUSE: 'BA Warehouse',
    cte.WAREHOUSE_REFRIGERATED: 'BA Warehouse',
    cte.SPORTS_LOCATION: 'BA Exercise Center',
    cte.SPORTS_ARENA: 'BA Sports Arena',
    cte.GYMNASIUM: 'BA Gymnasium',
    cte.MOTION_PICTURE_THEATRE: 'BA Motion Picture Theater',
    cte.MUSEUM: 'BA Museum',
    cte.PERFORMING_ARTS_THEATRE: 'BA Performing Arts Theater',
    cte.TRANSPORTATION: 'BA Transportation',
    cte.AUTOMOTIVE_FACILITY: 'BA Automotive Facility',
    cte.PARKING_GARAGE: 'BA Parking Garage',
    cte.RELIGIOUS: 'BA Religious Building',
    cte.NON_HEATED: 'n/a'
  }

  _usage_to_nrcan_usage = {
    cte.RESIDENTIAL: 'Multi-unit residential building',
    cte.SINGLE_FAMILY_HOUSE: 'Multi-unit residential building',
    cte.MULTI_FAMILY_HOUSE: 'Multi-unit residential building',
    cte.ROW_HOUSE: 'Multi-unit residential building',
    cte.MID_RISE_APARTMENT: 'Multi-unit residential building',
    cte.HIGH_RISE_APARTMENT: 'Multi-unit residential building',
    cte.OFFICE_AND_ADMINISTRATION: 'Office',
    cte.SMALL_OFFICE: 'Office',
    cte.MEDIUM_OFFICE: 'Office',
    cte.LARGE_OFFICE: 'Office',
    cte.COURTHOUSE: 'Courthouse',
    cte.FIRE_STATION: 'Fire station',
    cte.PENITENTIARY: 'Penitentiary',
    cte.POLICE_STATION: 'Police station',
    cte.POST_OFFICE: 'Post office',
    cte.LIBRARY: 'Library',
    cte.EDUCATION: 'School/university',
    cte.PRIMARY_SCHOOL: 'School/university',
    cte.PRIMARY_SCHOOL_WITH_SHOWER: 'School/university',
    cte.SECONDARY_SCHOOL: 'School/university',
    cte.UNIVERSITY: 'School/university',
    cte.LABORATORY_AND_RESEARCH_CENTER: 'School/university',
    cte.STAND_ALONE_RETAIL: 'Retail',
    cte.HOSPITAL: 'Hospital',
    cte.OUT_PATIENT_HEALTH_CARE: 'Health-care clinic',
    cte.HEALTH_CARE: 'Health-care clinic',
    cte.RETIREMENT_HOME_OR_ORPHANAGE: 'Health-care clinic',
    cte.COMMERCIAL: 'Retail',
    cte.STRIP_MALL: 'Retail',
    cte.SUPERMARKET: 'Retail',
    cte.RETAIL_SHOP_WITHOUT_REFRIGERATED_FOOD: 'Retail',
    cte.RETAIL_SHOP_WITH_REFRIGERATED_FOOD: 'Retail',
    cte.RESTAURANT: 'Dining - bar/lounge',
    cte.QUICK_SERVICE_RESTAURANT: 'Dining - cafeteria',
    cte.FULL_SERVICE_RESTAURANT: 'Dining - bar/lounge',
    cte.HOTEL: 'Hotel',
    cte.HOTEL_MEDIUM_CLASS: 'Motel',
    cte.SMALL_HOTEL: 'Motel',
    cte.LARGE_HOTEL: 'Hotel',
    cte.DORMITORY: 'Dormitory',
    cte.EVENT_LOCATION: 'Convention centre',
    cte.CONVENTION_CENTER: 'Convention centre',
    cte.HALL: 'Town hall',
    cte.GREEN_HOUSE: 'n/a',
    cte.INDUSTRY: 'Manufacturing facility',
    cte.WORKSHOP: 'Workshop',
    cte.WAREHOUSE: 'Warehouse',
    cte.WAREHOUSE_REFRIGERATED: 'Warehouse - refrigerated',
    cte.SPORTS_LOCATION: 'Exercise centre',
    cte.SPORTS_ARENA: 'Sports arena',
    cte.GYMNASIUM: 'Gymnasium',
    cte.MOTION_PICTURE_THEATRE: 'Motion picture theatre',
    cte.MUSEUM: 'Museum',
    cte.PERFORMING_ARTS_THEATRE: 'Performing arts theatre',
    cte.TRANSPORTATION: 'Transportation',
    cte.AUTOMOTIVE_FACILITY: 'Automotive facility',
    cte.PARKING_GARAGE: 'Parking garage',
    cte.RELIGIOUS: 'Religious',
    cte.NON_HEATED: 'n/a'
  }

  _function_to_nrcan_construction_function = {
    cte.RESIDENTIAL: 'MidriseApartment',
    cte.SINGLE_FAMILY_HOUSE: 'MidriseApartment',
    cte.MULTI_FAMILY_HOUSE: 'HighriseApartment',
    cte.ROW_HOUSE: 'MidriseApartment',
    cte.MID_RISE_APARTMENT: 'MidriseApartment',
    cte.HIGH_RISE_APARTMENT: 'HighriseApartment',
    cte.OFFICE_AND_ADMINISTRATION: 'MediumOffice',
    cte.SMALL_OFFICE: 'SmallOffice',
    cte.MEDIUM_OFFICE: 'MediumOffice',
    cte.LARGE_OFFICE: 'LargeOffice',
    cte.COURTHOUSE: 'MediumOffice',
    cte.FIRE_STATION: 'n/a',
    cte.PENITENTIARY: 'LargeHotel',
    cte.POLICE_STATION: 'n/a',
    cte.POST_OFFICE: 'MediumOffice',
    cte.LIBRARY: 'MediumOffice',
    cte.EDUCATION: 'SecondarySchool',
    cte.PRIMARY_SCHOOL: 'PrimarySchool',
    cte.PRIMARY_SCHOOL_WITH_SHOWER: 'PrimarySchool',
    cte.SECONDARY_SCHOOL: 'SecondarySchool',
    cte.UNIVERSITY: 'SecondarySchool',
    cte.LABORATORY_AND_RESEARCH_CENTER: 'SecondarySchool',
    cte.STAND_ALONE_RETAIL: 'RetailStandalone',
    cte.HOSPITAL: 'Hospital',
    cte.OUT_PATIENT_HEALTH_CARE: 'Outpatient',
    cte.HEALTH_CARE: 'Outpatient',
    cte.RETIREMENT_HOME_OR_ORPHANAGE: 'SmallHotel',
    cte.COMMERCIAL: 'RetailStripmall',
    cte.STRIP_MALL: 'RetailStripmall',
    cte.SUPERMARKET: 'RetailStripmall',
    cte.RETAIL_SHOP_WITHOUT_REFRIGERATED_FOOD: 'RetailStandalone',
    cte.RETAIL_SHOP_WITH_REFRIGERATED_FOOD: 'RetailStandalone',
    cte.RESTAURANT: 'FullServiceRestaurant',
    cte.QUICK_SERVICE_RESTAURANT: 'QuickServiceRestaurant',
    cte.FULL_SERVICE_RESTAURANT: 'FullServiceRestaurant',
    cte.HOTEL: 'SmallHotel',
    cte.HOTEL_MEDIUM_CLASS: 'SmallHotel',
    cte.SMALL_HOTEL: 'SmallHotel',
    cte.LARGE_HOTEL: 'LargeHotel',
    cte.DORMITORY: 'SmallHotel',
    cte.EVENT_LOCATION: 'n/a',
    cte.CONVENTION_CENTER: 'n/a',
    cte.HALL: 'n/a',
    cte.GREEN_HOUSE: 'n/a',
    cte.INDUSTRY: 'n/a',
    cte.WORKSHOP: 'n/a',
    cte.WAREHOUSE: 'Warehouse',
    cte.WAREHOUSE_REFRIGERATED: 'Warehouse',
    cte.SPORTS_LOCATION: 'n/a',
    cte.SPORTS_ARENA: 'n/a',
    cte.GYMNASIUM: 'n/a',
    cte.MOTION_PICTURE_THEATRE: 'n/a',
    cte.MUSEUM: 'n/a',
    cte.PERFORMING_ARTS_THEATRE: 'n/a',
    cte.TRANSPORTATION: 'n/a',
    cte.AUTOMOTIVE_FACILITY: 'n/a',
    cte.PARKING_GARAGE: 'n/a',
    cte.RELIGIOUS: 'n/a',
    cte.NON_HEATED: 'n/a'
  }

  _function_to_nrel_construction_function = {
    cte.RESIDENTIAL: 'residential',
    cte.SINGLE_FAMILY_HOUSE: 'residential',
    cte.MULTI_FAMILY_HOUSE: 'midrise apartment',
    cte.ROW_HOUSE: 'midrise apartment',
    cte.MID_RISE_APARTMENT: 'midrise apartment',
    cte.HIGH_RISE_APARTMENT: 'high-rise apartment',
    cte.OFFICE_AND_ADMINISTRATION: 'medium office',
    cte.SMALL_OFFICE: 'small office',
    cte.MEDIUM_OFFICE: 'medium office',
    cte.LARGE_OFFICE: 'large office',
    cte.COURTHOUSE: 'medium office',
    cte.FIRE_STATION: 'n/a',
    cte.PENITENTIARY: 'large hotel',
    cte.POLICE_STATION: 'n/a',
    cte.POST_OFFICE: 'medium office',
    cte.LIBRARY: 'medium office',
    cte.EDUCATION: 'secondary school',
    cte.PRIMARY_SCHOOL: 'primary school',
    cte.PRIMARY_SCHOOL_WITH_SHOWER: 'primary school',
    cte.SECONDARY_SCHOOL: 'secondary school',
    cte.UNIVERSITY: 'secondary school',
    cte.LABORATORY_AND_RESEARCH_CENTER: 'secondary school',
    cte.STAND_ALONE_RETAIL: 'stand-alone retail',
    cte.HOSPITAL: 'hospital',
    cte.OUT_PATIENT_HEALTH_CARE: 'outpatient healthcare',
    cte.HEALTH_CARE: 'outpatient healthcare',
    cte.RETIREMENT_HOME_OR_ORPHANAGE: 'small hotel',
    cte.COMMERCIAL: 'strip mall',
    cte.STRIP_MALL: 'strip mall',
    cte.SUPERMARKET: 'supermarket',
    cte.RETAIL_SHOP_WITHOUT_REFRIGERATED_FOOD: 'stand-alone retail',
    cte.RETAIL_SHOP_WITH_REFRIGERATED_FOOD: 'stand-alone retail',
    cte.RESTAURANT: 'full service restaurant',
    cte.QUICK_SERVICE_RESTAURANT: 'quick service restaurant',
    cte.FULL_SERVICE_RESTAURANT: 'full service restaurant',
    cte.HOTEL: 'small hotel',
    cte.HOTEL_MEDIUM_CLASS: 'small hotel',
    cte.SMALL_HOTEL: 'small hotel',
    cte.LARGE_HOTEL: 'large hotel',
    cte.DORMITORY: 'small hotel',
    cte.EVENT_LOCATION: 'n/a',
    cte.CONVENTION_CENTER: 'n/a',
    cte.HALL: 'n/a',
    cte.GREEN_HOUSE: 'n/a',
    cte.INDUSTRY: 'n/a',
    cte.WORKSHOP: 'n/a',
    cte.WAREHOUSE: 'warehouse',
    cte.WAREHOUSE_REFRIGERATED: 'warehouse',
    cte.SPORTS_LOCATION: 'n/a',
    cte.SPORTS_ARENA: 'n/a',
    cte.GYMNASIUM: 'n/a',
    cte.MOTION_PICTURE_THEATRE: 'n/a',
    cte.MUSEUM: 'n/a',
    cte.PERFORMING_ARTS_THEATRE: 'n/a',
    cte.TRANSPORTATION: 'n/a',
    cte.AUTOMOTIVE_FACILITY: 'n/a',
    cte.PARKING_GARAGE: 'n/a',
    cte.RELIGIOUS: 'n/a',
    cte.NON_HEATED: 'n/a'
  }

  _pluto_function_to_hub_function = {
    'A0': cte.SINGLE_FAMILY_HOUSE,
    'A1': cte.SINGLE_FAMILY_HOUSE,
    'A2': cte.SINGLE_FAMILY_HOUSE,
    'A3': cte.SINGLE_FAMILY_HOUSE,
    'A4': cte.SINGLE_FAMILY_HOUSE,
    'A5': cte.SINGLE_FAMILY_HOUSE,
    'A6': cte.SINGLE_FAMILY_HOUSE,
    'A7': cte.SINGLE_FAMILY_HOUSE,
    'A8': cte.SINGLE_FAMILY_HOUSE,
    'A9': cte.SINGLE_FAMILY_HOUSE,
    'B1': cte.MULTI_FAMILY_HOUSE,
    'B2': cte.MULTI_FAMILY_HOUSE,
    'B3': cte.MULTI_FAMILY_HOUSE,
    'B9': cte.MULTI_FAMILY_HOUSE,
    'C0': cte.RESIDENTIAL,
    'C1': cte.RESIDENTIAL,
    'C2': cte.RESIDENTIAL,
    'C3': cte.RESIDENTIAL,
    'C4': cte.RESIDENTIAL,
    'C5': cte.RESIDENTIAL,
    'C6': cte.RESIDENTIAL,
    'C7': cte.RESIDENTIAL,
    'C8': cte.RESIDENTIAL,
    'C9': cte.RESIDENTIAL,
    'D0': cte.RESIDENTIAL,
    'D1': cte.RESIDENTIAL,
    'D2': cte.RESIDENTIAL,
    'D3': cte.RESIDENTIAL,
    'D4': cte.RESIDENTIAL,
    'D5': cte.RESIDENTIAL,
    'D6': cte.RESIDENTIAL,
    'D7': cte.RESIDENTIAL,
    'D8': cte.RESIDENTIAL,
    'D9': cte.RESIDENTIAL,
    'E1': cte.WAREHOUSE,
    'E3': cte.WAREHOUSE,
    'E4': cte.WAREHOUSE,
    'E5': cte.WAREHOUSE,
    'E7': cte.WAREHOUSE,
    'E9': cte.WAREHOUSE,
    'F1': cte.WAREHOUSE,
    'F2': cte.WAREHOUSE,
    'F4': cte.WAREHOUSE,
    'F5': cte.WAREHOUSE,
    'F8': cte.WAREHOUSE,
    'F9': cte.WAREHOUSE,
    'G0': cte.SMALL_OFFICE,
    'G1': cte.SMALL_OFFICE,
    'G2': cte.SMALL_OFFICE,
    'G3': cte.SMALL_OFFICE,
    'G4': cte.SMALL_OFFICE,
    'G5': cte.SMALL_OFFICE,
    'G6': cte.SMALL_OFFICE,
    'G7': cte.SMALL_OFFICE,
    'G8': cte.SMALL_OFFICE,
    'G9': cte.SMALL_OFFICE,
    'H1': cte.HOTEL,
    'H2': cte.HOTEL,
    'H3': cte.HOTEL,
    'H4': cte.HOTEL,
    'H5': cte.HOTEL,
    'H6': cte.HOTEL,
    'H7': cte.HOTEL,
    'H8': cte.HOTEL,
    'H9': cte.HOTEL,
    'HB': cte.HOTEL,
    'HH': cte.HOTEL,
    'HR': cte.HOTEL,
    'HS': cte.HOTEL,
    'I1': cte.HOSPITAL,
    'I2': cte.OUT_PATIENT_HEALTH_CARE,
    'I3': cte.OUT_PATIENT_HEALTH_CARE,
    'I4': cte.RESIDENTIAL,
    'I5': cte.OUT_PATIENT_HEALTH_CARE,
    'I6': cte.OUT_PATIENT_HEALTH_CARE,
    'I7': cte.OUT_PATIENT_HEALTH_CARE,
    'I9': cte.OUT_PATIENT_HEALTH_CARE,
    'J1': cte.LARGE_OFFICE,
    'J2': cte.LARGE_OFFICE,
    'J3': cte.LARGE_OFFICE,
    'J4': cte.LARGE_OFFICE,
    'J5': cte.LARGE_OFFICE,
    'J6': cte.LARGE_OFFICE,
    'J7': cte.LARGE_OFFICE,
    'J8': cte.LARGE_OFFICE,
    'J9': cte.LARGE_OFFICE,
    'K1': cte.STRIP_MALL,
    'K2': cte.STRIP_MALL,
    'K3': cte.STRIP_MALL,
    'K4': cte.RESIDENTIAL,
    'K5': cte.RESTAURANT,
    'K6': cte.SUPERMARKET,
    'K7': cte.SUPERMARKET,
    'K8': cte.SUPERMARKET,
    'K9': cte.SUPERMARKET,
    'L1': cte.RESIDENTIAL,
    'L2': cte.RESIDENTIAL,
    'L3': cte.RESIDENTIAL,
    'L8': cte.RESIDENTIAL,
    'L9': cte.RESIDENTIAL,
    'M1': cte.LARGE_OFFICE,
    'M2': cte.LARGE_OFFICE,
    'M3': cte.LARGE_OFFICE,
    'M4': cte.LARGE_OFFICE,
    'M9': cte.LARGE_OFFICE,
    'N1': cte.RESIDENTIAL,
    'N2': cte.RESIDENTIAL,
    'N3': cte.RESIDENTIAL,
    'N4': cte.RESIDENTIAL,
    'N9': cte.RESIDENTIAL,
    'O1': cte.SMALL_OFFICE,
    'O2': cte.SMALL_OFFICE,
    'O3': cte.SMALL_OFFICE,
    'O4': cte.SMALL_OFFICE,
    'O5': cte.SMALL_OFFICE,
    'O6': cte.SMALL_OFFICE,
    'O7': cte.SMALL_OFFICE,
    'O8': cte.SMALL_OFFICE,
    'O9': cte.SMALL_OFFICE,
    'P1': cte.LARGE_OFFICE,
    'P2': cte.HOTEL,
    'P3': cte.SMALL_OFFICE,
    'P4': cte.SMALL_OFFICE,
    'P5': cte.SMALL_OFFICE,
    'P6': cte.SMALL_OFFICE,
    'P7': cte.LARGE_OFFICE,
    'P8': cte.LARGE_OFFICE,
    'P9': cte.SMALL_OFFICE,
    'Q0': cte.SMALL_OFFICE,
    'Q1': cte.SMALL_OFFICE,
    'Q2': cte.SMALL_OFFICE,
    'Q3': cte.SMALL_OFFICE,
    'Q4': cte.SMALL_OFFICE,
    'Q5': cte.SMALL_OFFICE,
    'Q6': cte.SMALL_OFFICE,
    'Q7': cte.SMALL_OFFICE,
    'Q8': cte.SMALL_OFFICE,
    'Q9': cte.SMALL_OFFICE,
    'R0': cte.RESIDENTIAL,
    'R1': cte.RESIDENTIAL,
    'R2': cte.RESIDENTIAL,
    'R3': cte.RESIDENTIAL,
    'R4': cte.RESIDENTIAL,
    'R5': cte.RESIDENTIAL,
    'R6': cte.RESIDENTIAL,
    'R7': cte.RESIDENTIAL,
    'R8': cte.RESIDENTIAL,
    'R9': cte.RESIDENTIAL,
    'RA': cte.RESIDENTIAL,
    'RB': cte.RESIDENTIAL,
    'RC': cte.RESIDENTIAL,
    'RD': cte.RESIDENTIAL,
    'RG': cte.RESIDENTIAL,
    'RH': cte.RESIDENTIAL,
    'RI': cte.RESIDENTIAL,
    'RK': cte.RESIDENTIAL,
    'RM': cte.RESIDENTIAL,
    'RR': cte.RESIDENTIAL,
    'RS': cte.RESIDENTIAL,
    'RW': cte.RESIDENTIAL,
    'RX': cte.RESIDENTIAL,
    'RZ': cte.RESIDENTIAL,
    'S0': cte.RESIDENTIAL,
    'S1': cte.RESIDENTIAL,
    'S2': cte.RESIDENTIAL,
    'S3': cte.RESIDENTIAL,
    'S4': cte.RESIDENTIAL,
    'S5': cte.RESIDENTIAL,
    'S9': cte.RESIDENTIAL,
    'U0': cte.WAREHOUSE,
    'U1': cte.WAREHOUSE,
    'U2': cte.WAREHOUSE,
    'U3': cte.WAREHOUSE,
    'U4': cte.WAREHOUSE,
    'U5': cte.WAREHOUSE,
    'U6': cte.WAREHOUSE,
    'U7': cte.WAREHOUSE,
    'U8': cte.WAREHOUSE,
    'U9': cte.WAREHOUSE,
    'W1': cte.PRIMARY_SCHOOL,
    'W2': cte.PRIMARY_SCHOOL,
    'W3': cte.SECONDARY_SCHOOL,
    'W4': cte.EDUCATION,
    'W5': cte.SECONDARY_SCHOOL,
    'W6': cte.SECONDARY_SCHOOL,
    'W7': cte.SECONDARY_SCHOOL,
    'W8': cte.PRIMARY_SCHOOL,
    'W9': cte.SECONDARY_SCHOOL,
    'Y1': cte.LARGE_OFFICE,
    'Y2': cte.LARGE_OFFICE,
    'Y3': cte.LARGE_OFFICE,
    'Y4': cte.LARGE_OFFICE,
    'Y5': cte.LARGE_OFFICE,
    'Y6': cte.LARGE_OFFICE,
    'Y7': cte.LARGE_OFFICE,
    'Y8': cte.LARGE_OFFICE,
    'Y9': cte.LARGE_OFFICE,
    'Z1': cte.LARGE_OFFICE
  }
  _hft_function_to_hub_function = {
    'residential': cte.RESIDENTIAL,
    'single family house': cte.SINGLE_FAMILY_HOUSE,
    'multifamily house': cte.MULTI_FAMILY_HOUSE,
    'hotel': cte.HOTEL,
    'hospital': cte.HOSPITAL,
    'outpatient': cte.OUT_PATIENT_HEALTH_CARE,
    'commercial': cte.SUPERMARKET,
    'strip mall': cte.STRIP_MALL,
    'warehouse': cte.WAREHOUSE,
    'primary school': cte.PRIMARY_SCHOOL,
    'secondary school': cte.EDUCATION,
    'office': cte.MEDIUM_OFFICE,
    'large office': cte.LARGE_OFFICE
  }

  @property
  def hub_usage_to_hft_usage(self):
    """
    Get HfT usage from the given internal usage
    :return: dict
    """
    return self._usage_to_hft_usage

  @property
  def hub_usage_to_comnet_usage(self):
    """
    Get Comnet usage from the given internal usage
    :return: dict
    """
    return self._usage_to_comnet_usage

  @property
  def hub_usage_to_nrcan_usage(self):
    """
    Get Nrcan usage from the given internal usage
    :return: dict
    """
    return self._usage_to_nrcan_usage

  @property
  def hub_function_to_nrcan_construction_function(self):
    """
    Get Nrcan construction function from the given internal function
    :return: dict
    """
    return self._function_to_nrcan_construction_function

  @property
  def hub_function_to_nrel_construction_function(self):
    """
    Get Nrel construction function from the given internal function
    :return: dict
    """
    return self._function_to_nrel_construction_function

  @property
  def pluto_function_to_hub_function(self):
    """
    Set internal function from pluto standard
    :return: dict
    """
    return self._pluto_function_to_hub_function

  @property
  def hft_function_to_hub_function(self):
    """
    Set internal function from functions used in hft files
    :return: dict
    """
    return self._hft_function_to_hub_function



"""
DatWeatherParameters class to extract weather parameters from a defined region in .dat format
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""

import sys
from pathlib import Path
import pandas as pd
import hub.helpers.constants as cte


class DatWeatherParameters:
  """
  DatWeatherParameters class
  """

  def __init__(self, city, path):
    self._weather_values = None
    self._city = city
    self._city_name = city.climate_reference_city
    file_name = 'inseldb_' + self._city_name + '.dat'
    self._path = Path(path / file_name)

    if self._weather_values is None:
      try:
        self._weather_values = pd.read_csv(self._path, sep=r'\s+', header=None,
                                           names=['hour', 'global_horiz', 'temperature', 'diffuse', 'beam', 'empty'])
      except SystemExit:
        sys.stderr.write(f'Error: weather file {self._path} not found\n')
        sys.exit()

    for building in self._city.buildings:
      new_value = pd.DataFrame(self._weather_values[['temperature']].to_numpy(), columns=['inseldb'])
      if cte.HOUR not in building.external_temperature:
        building.external_temperature[cte.HOUR] = new_value
      else:
        pd.concat([building.external_temperature[cte.HOUR], new_value], axis=1)

      new_value = pd.DataFrame(self._weather_values[['global_horiz']].to_numpy(), columns=['inseldb'])
      if cte.HOUR not in building.global_horizontal:
        building.global_horizontal[cte.HOUR] = new_value
      else:
        pd.concat([building.global_horizontal[cte.HOUR], new_value], axis=1)

      new_value = pd.DataFrame(self._weather_values[['diffuse']].to_numpy(), columns=['inseldb'])
      if cte.HOUR not in building.diffuse:
        building.diffuse[cte.HOUR] = new_value
      else:
        pd.concat([building.diffuse[cte.HOUR], new_value], axis=1)

      new_value = pd.DataFrame(self._weather_values[['beam']].to_numpy(), columns=['inseldb'])
      if cte.HOUR not in building.beam:
        building.beam[cte.HOUR] = new_value
      else:
        pd.concat([building.beam[cte.HOUR], new_value], axis=1)



"""
InselMonthlyEnergyBalance exports models to insel format
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""

import numpy as np
from pathlib import Path

from hub.exports.formats.insel import Insel
from hub.imports.weather.helpers.weather import Weather
import hub.helpers.constants as cte

_CONSTRUCTION_CODE = {
    cte.WALL: '1',
    cte.GROUND: '2',
    cte.ROOF: '3',
    cte.INTERIOR_WALL: '5',
    cte.GROUND_WALL: '6',
    cte.ATTIC_FLOOR: '7',
    cte.INTERIOR_SLAB: '8'
  }


class InselMonthlyEnergyBalance(Insel):

  def __init__(self, city, path, radiation_calculation_method='sra', weather_format='epw'):
    super().__init__(city, path)
    self._radiation_calculation_method = radiation_calculation_method
    self._weather_format = weather_format
    self._contents = []
    self._insel_files_paths = []
    for building in city.buildings:
      self._insel_files_paths.append(building.name + '.insel')
      file_name_out = building.name + '.out'
      output_path = Path(self._path / file_name_out).resolve()
      self._contents.append(self.generate_meb_template(building, output_path, self._radiation_calculation_method,
                                                       self._weather_format))
    self._export()

  def _export(self):
    for i_file, content in enumerate(self._contents):
      file_name = self._insel_files_paths[i_file]
      with open(Path(self._path / file_name).resolve(), 'w') as insel_file:
        insel_file.write(content)
    return

  @staticmethod
  def generate_meb_template(building, insel_outputs_path, radiation_calculation_method, weather_format):
    file = ""
    i_block = 1
    parameters = ["1", "12", "1"]
    file = Insel._add_block(file, i_block, 'DO', parameters=parameters)

    i_block = 4
    inputs = ["1.1", "20.1", "21.1"]
    surfaces = building.surfaces
    for i in range(1, len(surfaces) + 1):
      inputs.append(f"{str(100 + i)}.1 % Radiation surface {str(i)}")

    # BUILDING PARAMETERS
    parameters = [f'{0.85 * building.volume} % BP(1) Heated Volume (m3)',
                  f'{building.average_storey_height} % BP(2) Average storey height (m)',
                  f'{building.storeys_above_ground} % BP(3) Number of storeys above ground',
                  f'{building.attic_heated} % BP(4) Attic heating type (0=no room, 1=unheated, 2=heated)',
                  f'{building.basement_heated} % BP(5) Cellar heating type (0=no room, 1=unheated, 2=heated, '
                  f'99=invalid)']

    # todo: this method and the insel model have to be reviewed for more than one internal zone
    internal_zone = building.internal_zones[0]
    thermal_zone = internal_zone.thermal_zones[0]
    parameters.append(f'{thermal_zone.indirectly_heated_area_ratio} % BP(6) Indirectly heated area ratio')
    parameters.append(f'{thermal_zone.effective_thermal_capacity / 3600} % BP(7) Effective heat capacity (Wh/m2K)')
    parameters.append(f'{thermal_zone.additional_thermal_bridge_u_value} '
                      f'% BP(8) Additional U-value for heat bridge (Wh/m2K)')
    parameters.append('1 % BP(9) Usage type (0=standard, 1=IWU)')

    # ZONES AND SURFACES
    parameters.append(f'{len(internal_zone.usages)} %  BP(10) Number of zones')

    for i, usage in enumerate(internal_zone.usages):
      percentage_usage = usage.percentage
      parameters.append(f'{float(internal_zone.area) * percentage_usage}  % BP(11) #1 Area of zone {i + 1} (m2)')
      total_internal_gain = 0
      for ig in usage.internal_gains:
        total_internal_gain += float(ig.average_internal_gain) * \
                                (float(ig.convective_fraction) + float(ig.radiative_fraction))
      parameters.append(f'{total_internal_gain} % BP(12) #2 Internal gains of zone {i + 1}')
      parameters.append(f'{usage.thermal_control.mean_heating_set_point} % BP(13) #3 Heating setpoint temperature '
                        f'zone {i + 1} (degree Celsius)')
      parameters.append(f'{usage.thermal_control.heating_set_back} % BP(14) #4 Heating setback temperature '
                        f'zone {i + 1} (degree Celsius)')
      parameters.append(f'{usage.thermal_control.mean_cooling_set_point} % BP(15) #5 Cooling setpoint temperature '
                        f'zone {i + 1} (degree Celsius)')
      parameters.append(f'{usage.hours_day} %  BP(16) #6 Usage hours per day zone {i + 1}')
      parameters.append(f'{usage.days_year} %  BP(17) #7 Usage days per year zone {i + 1}')
      parameters.append(f'{usage.mechanical_air_change} % BP(18) #8 Minimum air change rate zone {i + 1} (ACH)')

    parameters.append(f'{len(thermal_zone.thermal_boundaries)}  % Number of surfaces = BP(11+8z) \n'
                      f'% 1. Surface type (1=wall, 2=ground 3=roof, 4=flat roof)\n'
                      f'% 2. Areas above ground (m2)\n'
                      f'% 3. Areas below ground (m2)\n'
                      f'% 4. U-value (W/m2K)\n'
                      f'% 5. Window area (m2)\n'
                      f'% 6. Window frame fraction\n'
                      f'% 7. Window U-value (W/m2K)\n'
                      f'% 8. Window g-value\n'
                      f'% 9. Short-wave reflectance\n'
                      f'% #1     #2       #3      #4      #5     #6     #7     #8     #9\n')

    for thermal_boundary in thermal_zone.thermal_boundaries:
      type_code = _CONSTRUCTION_CODE[thermal_boundary.type]
      window_area = thermal_boundary.opaque_area * thermal_boundary.window_ratio / (1 - thermal_boundary.window_ratio)

      parameters.append(type_code)
      if thermal_boundary.type != cte.GROUND:
        parameters.append(thermal_boundary.opaque_area + window_area)
        parameters.append('0.0')
      else:
        parameters.append('0.0')
        parameters.append(thermal_boundary.opaque_area + window_area)
      parameters.append(thermal_boundary.u_value)
      parameters.append(window_area)

      if window_area <= 0.001:
        parameters.append(0.0)
        parameters.append(0.0)
        parameters.append(0.0)
      else:
        thermal_opening = thermal_boundary.thermal_openings[0]
        parameters.append(thermal_opening.frame_ratio)
        parameters.append(thermal_opening.overall_u_value)
        parameters.append(thermal_opening.g_value)
      if thermal_boundary.type is not cte.GROUND:
        parameters.append(thermal_boundary.parent_surface.short_wave_reflectance)
      else:
        parameters.append(0.0)

    file = Insel._add_block(file, i_block, 'd18599', inputs=inputs, parameters=parameters)

    i_block = 20
    inputs = ['1']
    parameters = ['12 % Monthly ambient temperature (degree Celsius)']

    external_temperature = building.external_temperature[cte.MONTH]

    for i in range(0, len(external_temperature)):
      parameters.append(f'{i + 1} {external_temperature.at[i, weather_format]}')

    file = Insel._add_block(file, i_block, 'polyg', inputs=inputs, parameters=parameters)

    i_block = 21
    inputs = ['1']
    parameters = ['12 % Monthly sky temperature']

    sky_temperature = Weather.sky_temperature(external_temperature[[weather_format]].to_numpy().T[0])
    for i, temperature in enumerate(sky_temperature):
      parameters.append(f'{i + 1} {temperature}')

    file = Insel._add_block(file, i_block, 'polyg', inputs=inputs, parameters=parameters)

    for i, surface in enumerate(surfaces):
      i_block = 101 + i
      inputs = ['1 % Monthly surface radiation (W/m2)']
      parameters = [f'12 % Azimuth {np.rad2deg(surface.azimuth)}, '
                    f'inclination {np.rad2deg(surface.inclination)} (degrees)']

      if surface.type != 'Ground':
        global_irradiance = surface.global_irradiance[cte.MONTH]
        for j in range(0, len(global_irradiance)):
          parameters.append(f'{j + 1} {global_irradiance.at[j, radiation_calculation_method]}')
      else:
        for j in range(0, 12):
          parameters.append(f'{j + 1} 0.0')

      file = Insel._add_block(file, i_block, 'polyg', inputs=inputs, parameters=parameters)

    i_block = 300
    inputs = ['4.1', '4.2']
    file = Insel._add_block(file, i_block, 'cum', inputs=inputs)

    i_block = 303
    inputs = ['300.1', '300.2']
    file = Insel._add_block(file, i_block, 'atend', inputs=inputs)

    i_block = 310
    inputs = ['4.1', '4.2']
    parameters = ['1 % Mode',
                  '0 % Suppress FNQ inputs',
                  f"'{str(insel_outputs_path)}' % File name",
                  "'*' % Fortran format"]
    file = Insel._add_block(file, i_block, 'WRITE', inputs=inputs, parameters=parameters)

    return file



"""
Construction catalog Archetype
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Guille Gutierrez guillermo.gutierrezmorote@concordia.ca
"""

from hub.catalog_factories.data_models.construction.construction import Construction


class Archetype:
  def __init__(self, archetype_id,
               name,
               function,
               climate_zone,
               construction_period,
               constructions,
               average_storey_height,
               thermal_capacity,
               extra_loses_due_to_thermal_bridges,
               indirect_heated_ratio,
               infiltration_rate_for_ventilation_system_off,
               infiltration_rate_for_ventilation_system_on):
    self._id = archetype_id
    self._name = name
    self._function = function
    self._climate_zone = climate_zone
    self._construction_period = construction_period
    self._constructions = constructions
    self._average_storey_height = average_storey_height
    self._thermal_capacity = thermal_capacity
    self._extra_loses_due_to_thermal_bridges = extra_loses_due_to_thermal_bridges
    self._indirect_heated_ratio = indirect_heated_ratio
    self._infiltration_rate_for_ventilation_system_off = infiltration_rate_for_ventilation_system_off
    self._infiltration_rate_for_ventilation_system_on = infiltration_rate_for_ventilation_system_on

  @property
  def id(self):
    """
    Get archetype id
    :return: int
    """
    return self._id

  @property
  def name(self):
    """
    Get archetype name
    :return: str
    """
    return self._name

  @property
  def function(self):
    """
    Get archetype function
    :return: str
    """
    return self._function

  @property
  def climate_zone(self):
    """
    Get archetype climate zone
    :return: str
    """
    return self._climate_zone

  @property
  def constructions(self) -> [Construction]:
    """
    Get archetype constructions
    :return: [Construction]
    """
    return self._constructions

  @property
  def construction_period(self):
    """
    Get archetype construction period
    :return: str
    """
    return self._construction_period

  @property
  def average_storey_height(self):
    """
    Get archetype average storey height
    :return: float
    """
    return self._average_storey_height

  @property
  def thermal_capacity(self):
    """
    Get archetype thermal capacity
    :return: int
    """
    return self._thermal_capacity

  @property
  def extra_loses_due_to_thermal_bridges(self):
    """
    Get archetype extra loses due to thermal bridges
    :return: int
    """
    return self._extra_loses_due_to_thermal_bridges

  @property
  def indirect_heated_ratio(self):
    """
    Get archetype indirect heat ratio
    :return: float
    """
    return self._indirect_heated_ratio

  @property
  def infiltration_rate_for_ventilation_system_off(self):
    """
    Get archetype infiltration rate for ventilation system off
    :return: float
    """
    return self._infiltration_rate_for_ventilation_system_off

  @property
  def infiltration_rate_for_ventilation_system_on(self):
    """
    Get archetype infiltration rate for ventilation system on
    :return: float
    """
    return self._infiltration_rate_for_ventilation_system_on



"""
Sensor type module
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Guille Gutierrez guillermo.gutierrezmorote@concordia.ca
"""

from enum import Enum

class SensorType(Enum):
  HUMIDITY = 0
  TEMPERATURE = 1
  CO2 = 2
  NOISE = 3
  PRESSURE = 4
  DIRECT_RADIATION = 5
  DIFFUSE_RADIATION = 6
  GLOBAL_RADIATION = 7
  AIR_QUALITY = 8
  GAS_FLOW = 9
  ENERGY = 10



"""
monthly_to_hourly_demand module
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""
import calendar as cal
import pandas as pd
from hub.city_model_structure.building_demand.occupant import Occupant
import hub.helpers.constants as cte


class MonthlyToHourlyDemand:
  """
  MonthlyToHourlyDemand class
  """
  def __init__(self, building, conditioning_seasons):
    self._hourly_heating = pd.DataFrame()
    self._hourly_cooling = pd.DataFrame()
    self._building = building
    self._conditioning_seasons = conditioning_seasons

  def hourly_heating(self, key):
    """
    hourly distribution of the monthly heating of a building
    :param key: string
    :return: [hourly_heating]
    """
    # todo: this method and the insel model have to be reviewed for more than one thermal zone
    external_temp = self._building.external_temperature[cte.HOUR]
    # todo: review index depending on how the schedules are defined, either 8760 or 24 hours
    for usage in self._building.usages:
      temp_set = float(usage.heating_setpoint)-3
      temp_back = float(usage.heating_setback)-3
      # todo: if these are data frames, then they should be called as (Occupancy should be in low case):
      # usage.schedules.Occupancy
      # self._conditioning_seasons.heating
      occupancy = Occupant().get_complete_year_schedule(usage.schedules['Occupancy'])
      heating_schedule = self._conditioning_seasons['heating']

      hourly_heating = []
      i = 0
      j = 0
      temp_grad_day = []
      for month in range(1, 13):
        temp_grad_month = 0
        month_range = cal.monthrange(2015, month)[1]
        for _ in range(1, month_range+1):
          external_temp_med = 0
          for hour in range(0, 24):
            external_temp_med += external_temp[key][i]/24
          for hour in range(0, 24):
            if external_temp_med < temp_set and heating_schedule[month-1] == 1:
              if occupancy[hour] > 0:
                hdd = temp_set - external_temp[key][i]
                if hdd < 0:
                  hdd = 0
                temp_grad_day.append(hdd)
              else:
                hdd = temp_back - external_temp[key][i]
                if hdd < 0:
                  hdd = 0
                temp_grad_day.append(hdd)
            else:
              temp_grad_day.append(0)

            temp_grad_month += temp_grad_day[i]
            i += 1

        for _ in range(1, month_range + 1):
          for hour in range(0, 24):
            monthly_demand = self._building.heating[cte.MONTH][month-1]
            if monthly_demand == 'NaN':
              monthly_demand = 0
            if temp_grad_month == 0:
              hourly_demand = 0
            else:
              hourly_demand = float(monthly_demand)*float(temp_grad_day[j])/float(temp_grad_month)
            hourly_heating.append(hourly_demand)
            j += 1
      self._hourly_heating = pd.DataFrame(data=hourly_heating, columns=['monthly to hourly'])
    return self._hourly_heating

  def hourly_cooling(self, key):
    """
    hourly distribution of the monthly cooling of a building
    :param key: string
    :return: [hourly_cooling]
    """
    # todo: this method and the insel model have to be reviewed for more than one thermal zone
    external_temp = self._building.external_temperature[cte.HOUR]
    # todo: review index depending on how the schedules are defined, either 8760 or 24 hours
    for usage in self._building.usages:
      temp_set = float(usage.cooling_setpoint)
      temp_back = 100
      occupancy = Occupant().get_complete_year_schedule(usage.schedules['Occupancy'])
      cooling_schedule = self._conditioning_seasons['cooling']

      hourly_cooling = []
      i = 0
      j = 0
      temp_grad_day = []
      for month in range(1, 13):
        temp_grad_month = 0
        month_range = cal.monthrange(2015, month)[1]
        for _ in range(1, month_range[1] + 1):
          for hour in range(0, 24):
            if external_temp[key][i] > temp_set and cooling_schedule[month - 1] == 1:
              if occupancy[hour] > 0:
                cdd = external_temp[key][i] - temp_set
                if cdd < 0:
                  cdd = 0
                temp_grad_day.append(cdd)
              else:
                cdd = external_temp[key][i] - temp_back
                if cdd < 0:
                  cdd = 0
                temp_grad_day.append(cdd)
            else:
              temp_grad_day.append(0)

            temp_grad_month += temp_grad_day[i]
            i += 1

        for _ in range(1, month_range[1] + 1):
          for hour in range(0, 24):
            #            monthly_demand = self._building.heating[cte.MONTH]['INSEL'][month-1]
            monthly_demand = self._building.cooling[cte.MONTH][month - 1]
            if monthly_demand == 'NaN':
              monthly_demand = 0
            if temp_grad_month == 0:
              hourly_demand = 0
            else:
              hourly_demand = float(monthly_demand) * float(temp_grad_day[j]) / float(temp_grad_month)
            hourly_cooling.append(hourly_demand)
            j += 1
      self._hourly_cooling = pd.DataFrame(data=hourly_cooling, columns=['monthly to hourly'])
    return self._hourly_cooling



"""
User performs user-related crud operations
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project CoderPeter Yefi peteryefi@gmail.com
"""
from hub.persistence import UserRepo
from hub.persistence import UserRoles


class UserFactory:
  """
  UserFactory class
  """

  def __init__(self, db_name, app_env, dotenv_path):
    self._user_repo = UserRepo(db_name=db_name, app_env=app_env, dotenv_path=dotenv_path)

  def create_user(self, name: str, email: str, password: str, role: UserRoles):
    """
    Creates a new user
    :param name: the name of the user
    :param email: the email of the user
    :param password: the password of the user
    :param role: the role of the user
    """
    return self._user_repo.insert(name, email, password, role)

  def update_user(self, user_id: int, name: str, email: str, password: str, role: UserRoles):
    """
    Creates a new user
    :param user_id: the id of the user
    :param name: the name of the user
    :param email: the email of the user
    :param password: the password of the user
    :param role: the role of the user
    """
    return self._user_repo.update(user_id, name, email, password, role)

  def delete_user(self, user_id):
    """
    Retrieve a single user
    :param user_id: the id of the user to delete
    """
    return self._user_repo.delete_user(user_id)



from .repository import Repository
from .repositories.city import City
from .repositories.application import Application
from .repositories.simulation_results import SimulationResults
from .repositories.city_object import CityObject
from .db_setup import DBSetup
from .repositories.user import User
from .models.user import UserRoles



"""
Origin-Destination node module
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""
from typing import List, TypeVar

from hub.city_model_structure.attributes.node import Node
from hub.city_model_structure.attributes.point import Point
from hub.city_model_structure.attributes.polygon import Polygon
from hub.city_model_structure.city_object import CityObject

OriginDestinationEdge = TypeVar('OriginDestinationEdge')


class OriginDestinationNode(Node):
  """
  OriginDestinationNode class
  """

  def __init__(self, name, coordinates, node_type='OriginDestinationNode', edges=None, polygon=None):
    super().__init__(name, edges)
    self._coordinates = coordinates
    self._node_type = node_type
    self._polygon = polygon
    self._land_use_types = None
    self._city_objects = None

  @property
  def node_type(self):
    """
    Get node type
    :return: str
    """
    return self._node_type

  @property
  def coordinates(self) -> Point:
    """
    Get node coordinates
    :return: Point
    """
    return self._coordinates

  @coordinates.setter
  def coordinates(self, value):
    """
    Set node coordinates
    :param value: Point
    """
    self._coordinates = value

  @property
  def edges(self) -> List[OriginDestinationEdge]:
    """
    get edges delimited by the node
    :return: [OriginDestinationEdge]
    """
    return self._edges

  @property
  def polygon(self) -> Polygon:
    """
    Get node polygon that defines the zone represented by the node
    :return: Polygon
    """
    return self._polygon

  @property
  def land_use_types(self) -> dict:
    """
    Get land use types inside the node polygon. It returns a dictionary with the types of land use together with the
    percentage of the land that corresponds to each type
    :return: {string : float}
    """
    return self._land_use_types

  @property
  def city_objects(self) -> List[CityObject]:
    """
    Get the list of city objects place inside the zone
    :return: List[CityObject]
    """
    return self._city_objects



"""
Insel export models to insel format
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""

from abc import ABC


class Insel(ABC):
  def __init__(self, city, path):
    self._city = city
    self._path = path
    self._results = None

  @staticmethod
  def _add_block(file, block_number, block_type, inputs='', parameters=''):
    file += "S " + str(block_number) + " " + block_type + "\n"
    for block_input in inputs:
      file += str(block_input) + "\n"
    if len(parameters) > 0:
      file += "P " + str(block_number) + "\n"
    for block_parameter in parameters:
      file += str(block_parameter) + "\n"
    return file

  def _export(self):
    raise NotImplementedError



"""
Usage catalog internal gain
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Guille Gutierrez guillermo.gutierrezmorote@concordia.ca
"""


class InternalGain:
  """
  InternalGain class
  """

  def __init__(self, internal_gain_type, average_internal_gain, convective_fraction, radiative_fraction, latent_fraction, schedules):
    self._type = internal_gain_type
    self._average_internal_gain = average_internal_gain
    self._convective_fraction = convective_fraction
    self._radiative_fraction = radiative_fraction
    self._latent_fraction = latent_fraction
    self._schedules = schedules





"""
Model representation of a User
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Peter Yefi peteryefi@gmail.com
"""

import datetime
import enum

from sqlalchemy import Column, Integer, String, Sequence
from sqlalchemy import DateTime, Enum

from hub.persistence.configuration import Models


class UserRoles(enum.Enum):
  Admin = 'Admin'
  Hub_Reader = 'Hub_Reader'


class User(Models):
  """
  A model representation of a city
  """
  __tablename__ = 'user'
  id = Column(Integer, Sequence('user_id_seq'), primary_key=True)
  name = Column(String, nullable=False)
  password = Column(String, nullable=False)
  role = Column(Enum(UserRoles), nullable=False, default=UserRoles.Hub_Reader)
  application_id = Column(Integer, nullable=False)
  created = Column(DateTime, default=datetime.datetime.utcnow)
  updated = Column(DateTime, default=datetime.datetime.utcnow)

  def __init__(self, name, password, role, application_id):
    self.name = name
    self.password = password
    self.role = role
    self.application_id = application_id



"""
Usage catalog thermal control
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Guille Gutierrez guillermo.gutierrezmorote@concordia.ca
"""
from typing import Union, List
from hub.catalog_factories.data_models.usages.schedule import Schedule


class ThermalControl:
  """
  ThermalControl class
  """
  def __init__(self, mean_heating_set_point,
               heating_set_back,
               mean_cooling_set_point,
               hvac_availability_schedules,
               heating_set_point_schedules,
               cooling_set_point_schedules):
    self._mean_heating_set_point = mean_heating_set_point
    self._heating_set_back = heating_set_back
    self._mean_cooling_set_point = mean_cooling_set_point
    self._hvac_availability_schedules = hvac_availability_schedules
    self._heating_set_point_schedules = heating_set_point_schedules
    self._cooling_set_point_schedules = cooling_set_point_schedules

  @property
  def mean_heating_set_point(self) -> Union[None, float]:
    """
    Get heating set point defined for a thermal zone in Celsius
    :return: None or float
    """
    return self._mean_heating_set_point

  @property
  def heating_set_back(self) -> Union[None, float]:
    """
    Get heating set back defined for a thermal zone in Celsius
    :return: None or float
    """
    return self._heating_set_back

  @property
  def mean_cooling_set_point(self) -> Union[None, float]:
    """
    Get cooling set point defined for a thermal zone in Celsius
    :return: None or float
    """
    return self._mean_cooling_set_point

  @property
  def hvac_availability_schedules(self) -> Union[None, List[Schedule]]:
    """
    Get the availability of the conditioning system defined for a thermal zone
    dataType = on/off
    :return: None or [Schedule]
    """
    return self._hvac_availability_schedules

  @property
  def heating_set_point_schedules(self) -> Union[None, List[Schedule]]:
    """
    Get heating set point schedule defined for a thermal zone in Celsius
    dataType = temperature
    :return: None or [Schedule]
    """
    return self._heating_set_point_schedules

  @property
  def cooling_set_point_schedules(self) -> Union[None, List[Schedule]]:
    """
    Get cooling set point schedule defined for a thermal zone in Celsius
    dataType = temperature
    :return: None or [Schedule]
    """
    return self._cooling_set_point_schedules



"""
TestConstructionCatalog
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Guille Gutierrez guillermo.gutierrezmorote@concordia.ca
"""

from unittest import TestCase
from hub.catalog_factories.construction_catalog_factory import ConstructionCatalogFactory


class TestConstructionCatalog(TestCase):

  def test_nrel_catalog(self):
    catalog = ConstructionCatalogFactory('nrel').catalog
    catalog_categories = catalog.names()
    constructions = catalog.names('constructions')
    windows = catalog.names('windows')
    materials = catalog.names('materials')
    self.assertTrue(len(constructions['constructions']), 24)
    self.assertTrue(len(windows['windows']), 4)
    self.assertTrue(len(materials['materials']), 19)
    with self.assertRaises(ValueError):
      catalog.names('unknown')

    # retrieving all the entries should not raise any exceptions
    for category in catalog_categories:
      for value in catalog_categories[category]:
        catalog.get_entry(value)

    with self.assertRaises(IndexError):
      catalog.get_entry('unknown')




"""
Subway entrance module
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""
from hub.city_model_structure.city_object import CityObject


class SubwayEntrance(CityObject):
  """
  SubwayEntrance(CityObject) class
  """
  def __init__(self, name, latitude, longitude):
    super().__init__(name, 0)
    self._name = name
    self._latitude = latitude
    self._longitude = longitude
    self._type = 'subway_entrance'

  @property
  def latitude(self):
    # todo: to be defined the spacial point and the units
    """
    Get latitude
    :return: float
    """
    return self._latitude

  @property
  def longitude(self):
    # todo: to be defined the spacial point and the units
    """
    Get longitude
    :return: float
    """
    return self._longitude

  @property
  def name(self):
    """
    Get name
    :return: str
    """
    return self._name



from hub.helpers import constants as cte


class ConstructionHelper:
  """
  Construction helper class
  """
  _reference_standard_to_construction_period = {
    'non_standard_dompark': '1900 - 2004',
    'ASHRAE 90.1_2004': '2004 - 2009',
    'ASHRAE 189.1_2009': '2009 - PRESENT'
  }

  _nrel_surfaces_types_to_hub_types = {
    'exterior wall': cte.WALL,
    'interior wall': cte.INTERIOR_WALL,
    'ground wall': cte.GROUND_WALL,
    'exterior slab': cte.GROUND,
    'attic floor': cte.ATTIC_FLOOR,
    'interior slab': cte.INTERIOR_SLAB,
    'roof': cte.ROOF
  }

  _nrcan_surfaces_types_to_hub_types = {
    'Wall_Outdoors': cte.WALL,
    'RoofCeiling_Outdoors': cte.ROOF,
    'Floor_Outdoors': cte.ATTIC_FLOOR,
    'Window_Outdoors': cte.WINDOW,
    'Skylight_Outdoors': cte.SKYLIGHT,
    'Door_Outdoors': cte.DOOR,
    'Wall_Ground': cte.GROUND_WALL,
    'RoofCeiling_Ground': cte.GROUND_WALL,
    'Floor_Ground': cte.GROUND
  }

  @property
  def reference_standard_to_construction_period(self):
    return self._reference_standard_to_construction_period

  @property
  def nrel_surfaces_types_to_hub_types(self):
    return self._nrel_surfaces_types_to_hub_types

  @property
  def nrcan_surfaces_types_to_hub_types(self):
    return self._nrcan_surfaces_types_to_hub_types



"""
Application repository with database CRUD operations
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Guille Gutierrez Guillermo.GutierrezMorote@concordia.ca
"""

import datetime
from typing import Union, Dict

from sqlalchemy import select
from sqlalchemy.exc import SQLAlchemyError

from hub.hub_logger import logger
from hub.persistence import Repository
from hub.persistence.models import Application as Model


class Application(Repository):
  _instance = None

  def __init__(self, db_name: str, dotenv_path: str, app_env: str):
    super().__init__(db_name, dotenv_path, app_env)

  def __new__(cls, db_name, dotenv_path, app_env):
    """
    Implemented for a singleton pattern
    """
    if cls._instance is None:
      cls._instance = super(Application, cls).__new__(cls)
    return cls._instance

  def insert(self, name: str, description: str, application_uuid: str) -> Union[Model, Dict]:
    """
    Inserts a new application
    :param name: Application name
    :param description: Application description
    :param application_uuid: Unique identifier for the application
    :return: application and dictionary
    """
    application = self.get_by_uuid(application_uuid)
    if application is None:
      try:
        application = Model(name=name, description=description, application_uuid=application_uuid)
        self.session.add(application)
        self.session.commit()
        return application
      except SQLAlchemyError as err:
        logger.error(f'An error occurred while creating application: {err}')
    else:
      return {'message': f'An application with {application_uuid} application uuid, already exists'}

  def update(self, application_uuid: str, name: str, description: str) -> Union[Dict, None]:
    """
    Updates an application
    :param application_uuid: the application uuid of the application to be updated
    :param name: the application name
    :param description: the application description
    :return:
    """
    try:
      self.session.query(Model).filter(
        Model.application_uuid == application_uuid
      ).update({'name': name, 'description': description, 'updated': datetime.datetime.utcnow()})
      self.session.commit()
    except SQLAlchemyError as err:
      logger.error(f'Error while updating application: {err}')
      return {'message': 'Error occurred while updating application'}

  def delete(self, application_uuid: str):
    """
    Deletes an application with the application_uuid
    :param application_uuid: The application uuid
    :return: None
    """
    try:
      self.session.query(Model).filter(Model.application_uuid == application_uuid).delete()
      self.session.flush()
      self.session.commit()
    except SQLAlchemyError as err:
      logger.error(f'Error while deleting application: {err}')

  def get_by_uuid(self, application_uuid: str) -> Union[Model, None]:
    """
    Fetch Application based on the application uuid
    :param application_uuid: the application uuid
    :return: Application with the provided application_uuid or None
    """
    try:
      result_set = self.session.execute(select(Model).where(
        Model.application_uuid == application_uuid)
      ).first()
      if result_set is None:
        return None
      return result_set[0]
    except SQLAlchemyError as err:
      logger.error(f'Error while fetching application by application_uuid: {err}')




"""
Bus node module
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""

from typing import List, TypeVar

from hub.city_model_structure.attributes.node import Node
from hub.city_model_structure.attributes.point import Point

BusEdge = TypeVar('BusEdge')


class BusNode(Node):
  """
  BusNode class
  """

  def __init__(self, name, coordinates, node_type='BusNode', edges=None):
    super().__init__(name, edges)
    self._coordinates = coordinates
    self._node_type = node_type

  @property
  def node_type(self):
    """
    Get node type
    :return: str
    """
    return self._node_type

  @property
  def coordinates(self) -> Point:
    """
    Get node coordinates
    :return: Point
    """
    return self._coordinates

  @coordinates.setter
  def coordinates(self, value):
    """
    Set node coordinates
    :param value: Point
    """
    self._coordinates = value

  @property
  def edges(self) -> List[BusEdge]:
    """
    get edges delimited by the node
    :return: [BusEdge]
    """
    return self._edges



"""
XlsWeatherParameters class to extract weather parameters from a defined region in .xls format
defined and used by the norm iso 52016-1:2017
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""

import sys
from pathlib import Path
import pandas as pd
import hub.helpers.constants as cte


class XlsWeatherParameters:
  """
  XlsWeatherParameters class
  """
  def __init__(self, city, path, name):
    self._weather_values = None
    self._city = city
    self._file_name = name
    file_name = self._file_name + '.xls'
    self._path = Path(path / file_name)

    if self._weather_values is None:
      try:
        self._weather_values = pd.read_excel(self._path, skiprows=4, nrows=9504,
                                             names=['hours_calc', 'month', 'week', 'day_of_week', 'hours_of_day',
                                                    'temperature', 'I_sol_vertical_N', 'I_sol_vertical_E',
                                                    'I_sol_vertical_S', 'I_sol_vertical_W', 'I_sol_45_N', 'I_sol_45_S',
                                                    'void', 'global_horiz', 'wind_velocity', 'humidity'])
      except SystemExit:
        sys.stderr.write(f'Error reading weather file {self._path}\n')
        sys.exit()

    for building in self._city.buildings:
      new_value = pd.DataFrame(self._weather_values[['temperature']].to_numpy(), columns=['iso52016'])
      if cte.HOUR not in building.external_temperature:
        building.external_temperature[cte.HOUR] = new_value
      else:
        pd.concat([building.external_temperature[cte.HOUR], new_value], axis=1)

      new_value = pd.DataFrame(self._weather_values[['global_horiz']].to_numpy(), columns=['iso52016'])
      if cte.HOUR not in building.global_horizontal:
        building.global_horizontal[cte.HOUR] = new_value
      else:
        pd.concat([building.global_horizontal[cte.HOUR], new_value], axis=1)



"""
City module
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Guille Gutierrez guillermo.gutierrezmorote@concordia.ca
Code contributors: Peter Yefi peteryefi@gmail.com
"""
from __future__ import annotations
import sys
import pickle
import math
import copy
import pyproj
from typing import List, Union
from pyproj import Transformer
from pathlib import Path
from hub.city_model_structure.building import Building
from hub.city_model_structure.city_object import CityObject
from hub.city_model_structure.city_objects_cluster import CityObjectsCluster
from hub.city_model_structure.buildings_cluster import BuildingsCluster
from hub.city_model_structure.fuel import Fuel
from hub.city_model_structure.iot.station import Station
from hub.city_model_structure.level_of_detail import LevelOfDetail
from hub.city_model_structure.machine import Machine
from hub.city_model_structure.parts_consisting_building import PartsConsistingBuilding
from hub.city_model_structure.subway_entrance import SubwayEntrance
from hub.helpers.geometry_helper import GeometryHelper
from hub.helpers.location import Location
from hub.city_model_structure.energy_system import EnergySystem
from hub.city_model_structure.lca_material import LcaMaterial


class City:
  """
  City class
  """

  def __init__(self, lower_corner, upper_corner, srs_name):
    self._name = None
    self._lower_corner = lower_corner
    self._upper_corner = upper_corner
    self._buildings = None
    self._subway_entrances = None
    self._srs_name = srs_name
    # todo: right now extracted at city level, in the future should be extracted also at building level if exist
    self._location = None
    self._country_code = None
    self._climate_reference_city = None
    self._climate_file = None
    self._latitude = None
    self._longitude = None
    self._time_zone = None
    self._buildings_clusters = None
    self._parts_consisting_buildings = None
    self._city_objects_clusters = None
    self._city_objects = None
    self._energy_systems = None
    self._fuels = None
    self._machines = None
    self._stations = []
    self._lca_materials = None
    self._level_of_detail = LevelOfDetail()

  @property
  def fuels(self) -> [Fuel]:
    return self._fuels

  @fuels.setter
  def fuels(self, value):
    self._fuels = value

  @property
  def machines(self) -> [Machine]:
    return self._machines

  @machines.setter
  def machines(self, value):
    self._machines = value

  def _get_location(self) -> Location:
    if self._location is None:
      gps = pyproj.CRS('EPSG:4326')  # LatLon with WGS84 datum used by GPS units and Google Earth
      try:
        if self._srs_name in GeometryHelper.srs_transformations.keys():
          self._srs_name = GeometryHelper.srs_transformations[self._srs_name]
        input_reference = pyproj.CRS(self.srs_name)  # Projected coordinate system from input data
      except pyproj.exceptions.CRSError:
        sys.stderr.write('Invalid projection reference system, please check the input data. '
                         '(e.g. in CityGML files: srs_name)\n')
        sys.exit()
      transformer = Transformer.from_crs(input_reference, gps)
      coordinates = transformer.transform(self.lower_corner[0], self.lower_corner[1])
      self._location = GeometryHelper.get_location(coordinates[0], coordinates[1])
    return self._location

  @property
  def country_code(self):
    """
    Get models country code
    :return: str
    """
    return self._get_location().country

  @property
  def name(self):
    """
    Get city name
    :return: str
    """
    return self._get_location().city

  @property
  def climate_reference_city(self) -> Union[None, str]:
    """
    Get the name for the climatic information reference city
    :return: None or str
    """
    return self._climate_reference_city

  @climate_reference_city.setter
  def climate_reference_city(self, value):
    """
    Set the name for the climatic information reference city
    :param value: str
    """
    if value is not None:
      self._climate_reference_city = str(value)

  @property
  def climate_file(self) -> Union[None, Path]:
    """
    Get the climate file full path
    :return: None or Path
    """
    return self._climate_file

  @climate_file.setter
  def climate_file(self, value):
    """
    Set the climate file full path
    :param value: Path
    """
    if value is not None:
      self._climate_file = Path(value)

  @property
  def city_objects(self) -> Union[List[CityObject], None]:
    """
    Get the city objects belonging to the city
    :return: None or [CityObject]
    """
    if self._city_objects is None:
      if self.city_objects_clusters is None:
        self._city_objects = []
      else:
        self._city_objects = self.city_objects_clusters
      if self.buildings is not None:
        for building in self.buildings:
          self._city_objects.append(building)
      if self.subway_entrances is not None:
        for subway_entrance in self.subway_entrances:
          self._city_objects.append(subway_entrance)
      if self.energy_systems is not None:
        for energy_system in self.energy_systems:
          self._city_objects.append(energy_system)
    return self._city_objects

  @property
  def buildings(self) -> Union[List[Building], None]:
    """
    Get the buildings belonging to the city
    :return: None or [Building]
    """
    return self._buildings

  @property
  def subway_entrances(self) -> Union[List[SubwayEntrance], None]:
    """
    Get the subway entrances belonging to the city
    :return: a list of subway entrances objects or none
    """
    return self._subway_entrances

  @property
  def lower_corner(self) -> List[float]:
    """
    Get city lower corner
    :return: [x,y,z]
    """
    return self._lower_corner

  @property
  def upper_corner(self) -> List[float]:
    """
    Get city upper corner
    :return: [x,y,z]
    """
    return self._upper_corner

  def city_object(self, name) -> Union[CityObject, None]:
    """
    Retrieve the city CityObject with the given name
    :param name:str
    :return: None or CityObject
    """
    for city_object in self.buildings:
      if str(city_object.name) == str(name):
        return city_object
    return None

  def add_city_object(self, new_city_object):
    """
    Add a CityObject to the city
    :param new_city_object:CityObject
    :return: None or not implemented error
    """
    if new_city_object.type == 'building':
      if self._buildings is None:
        self._buildings = []
      self._buildings.append(new_city_object)
    elif new_city_object.type == 'subway_entrance':
      if self._subway_entrances is None:
        self._subway_entrances = []
      self._subway_entrances.append(new_city_object)
    elif new_city_object.type == 'energy_system':
      if self._energy_systems is None:
        self._energy_systems = []
      self._energy_systems.append(new_city_object)
    else:
      raise NotImplementedError(new_city_object.type)

  def remove_city_object(self, city_object):
    """
    Remove a CityObject from the city
    :param city_object:CityObject
    :return: None
    """
    if city_object.type != 'building':
      raise NotImplementedError(city_object.type)
    if self._buildings is None or self._buildings == []:
      sys.stderr.write('Warning: impossible to remove city_object, the city is empty\n')
    else:
      if city_object in self._buildings:
        self._buildings.remove(city_object)

  @property
  def srs_name(self) -> Union[None, str]:
    """
    Get city srs name
    :return: None or str
    """
    return self._srs_name

  @name.setter
  def name(self, value):
    """
    Set city name
    :param value:str
    """
    if value is not None:
      self._name = str(value)

  @staticmethod
  def load(city_filename) -> City:
    """
    Load a city saved with city.save(city_filename)
    :param city_filename: city filename
    :return: City
    """
    with open(city_filename, 'rb') as file:
      return pickle.load(file)

  def save(self, city_filename):
    """
    Save a city into the given filename
    :param city_filename: destination city filename
    :return: None
    """
    with open(city_filename, 'wb') as file:
      pickle.dump(self, file)

  def region(self, center, radius) -> City:
    """
    Get a region from the city
    :param center: specific point in space [x, y, z]
    :param radius: distance to center of the sphere selected in meters
    :return: selected_region_city
    """
    selected_region_lower_corner = [center[0] - radius, center[1] - radius, center[2] - radius]
    selected_region_upper_corner = [center[0] + radius, center[1] + radius, center[2] + radius]
    selected_region_city = City(selected_region_lower_corner, selected_region_upper_corner, srs_name=self.srs_name)
    selected_region_city.climate_file = self.climate_file
    #    selected_region_city.climate_reference_city = self.climate_reference_city
    for city_object in self.city_objects:
      location = city_object.centroid
      if location is not None:
        distance = math.sqrt(math.pow(location[0] - center[0], 2) + math.pow(location[1] - center[1], 2)
                             + math.pow(location[2] - center[2], 2))
        if distance < radius:
          selected_region_city.add_city_object(city_object)
    return selected_region_city

  @property
  def latitude(self) -> Union[None, float]:
    """
    Get city latitude in degrees
    :return: None or float
    """
    return self._latitude

  @latitude.setter
  def latitude(self, value):
    """
    Set city latitude in degrees
    :parameter value: float
    """
    if value is not None:
      self._latitude = float(value)

  @property
  def longitude(self) -> Union[None, float]:
    """
    Get city longitude in degrees
    :return: None or float
    """
    return self._longitude

  @longitude.setter
  def longitude(self, value):
    """
    Set city longitude in degrees
    :parameter value: float
    """
    if value is not None:
      self._longitude = float(value)

  @property
  def time_zone(self) -> Union[None, float]:
    """
    Get city time_zone
    :return: None or float
    """
    return self._time_zone

  @time_zone.setter
  def time_zone(self, value):
    """
    Set city time_zone
    :parameter value: float
    """
    if value is not None:
      self._time_zone = float(value)

  @property
  def buildings_clusters(self) -> Union[List[BuildingsCluster], None]:
    """
    Get buildings clusters belonging to the city
    :return: None or [BuildingsCluster]
    """
    return self._buildings_clusters

  @property
  def parts_consisting_buildings(self) -> Union[List[PartsConsistingBuilding], None]:
    """
    Get parts consisting buildings belonging to the city
    :return: None or [PartsConsistingBuilding]
    """
    return self._parts_consisting_buildings

  @property
  def energy_systems(self) -> Union[List[EnergySystem], None]:
    """
    Get energy systems belonging to the city
    :return: None or [EnergySystem]
    """
    return self._energy_systems

  @property
  def stations(self) -> [Station]:
    """
    Get the sensors stations belonging to the city
    :return: [Station]
    """
    return self._stations

  @property
  def city_objects_clusters(self) -> Union[List[CityObjectsCluster], None]:
    """
    Get city objects clusters belonging to the city
    :return: None or [CityObjectsCluster]
    """
    if self.buildings_clusters is None:
      self._city_objects_clusters = []
    else:
      self._city_objects_clusters = self.buildings_clusters
    if self.parts_consisting_buildings is not None:
      self._city_objects_clusters.append(self.parts_consisting_buildings)
    return self._city_objects_clusters

  def add_city_objects_cluster(self, new_city_objects_cluster):
    """
    Add a CityObject to the city
    :param new_city_objects_cluster:CityObjectsCluster
    :return: None or NotImplementedError
    """
    if new_city_objects_cluster.type == 'buildings':
      if self._buildings_clusters is None:
        self._buildings_clusters = []
      self._buildings_clusters.append(new_city_objects_cluster)
    elif new_city_objects_cluster.type == 'building_parts':
      if self._parts_consisting_buildings is None:
        self._parts_consisting_buildings = []
      self._parts_consisting_buildings.append(new_city_objects_cluster)
    else:
      raise NotImplementedError

  @property
  def lca_materials(self) -> Union[List[LcaMaterial], None]:
    """
    Get life cycle materials for the city
    :return: [LcaMaterial] or
    """
    return self._lca_materials

  @lca_materials.setter
  def lca_materials(self, value):
    """
    Set life cycle materials for the city
    """
    self._lca_materials = value

  def lca_material(self, lca_id) -> Union[LcaMaterial, None]:
    """
    Get the lca material matching the given Id
    :return: LcaMaterial or None
    """
    for lca_material in self.lca_materials:
      if str(lca_material.id) == str(lca_id):
        return lca_material
    return None

  @property
  def copy(self) -> City:
    """
    Get a copy of the current city
    """
    return copy.deepcopy(self)

  def merge(self, city) -> City:
    _merge_city = self.copy
    for city_object in city.city_objects:
      _merge_city.add_city_object(city_object)
    return _merge_city

  @property
  def level_of_detail(self) -> LevelOfDetail:
    """
    Get level of detail of different aspects of the city: geometry, construction and usage
    :return: LevelOfDetail
    """
    return self._level_of_detail



"""
Concordia energy consumption
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Guille Gutierrez guillermo.gutierrezmorote@concordia.ca
"""
import pandas as pd
from hub.imports.sensors.concordia_file_report import ConcordiaFileReport


class ConcordiaEnergyConsumption(ConcordiaFileReport):
  """
  Concordia energy consumption sensor class
  """
  def __init__(self, city, end_point, base_path):
    super().__init__(city, end_point, base_path, 'concordia_energy_db.json')
    for city_object in city.city_objects:
      self._assign_sensor_to_object(city_object)

  def _assign_sensor_to_object(self, obj):
    for i in range(len(self._city_object)):
      if self._city_object[i] == obj.name and self._sensors[i] in self._sensor_point:
        building_measures = [self._measures["Date time"], self._measures[self._sensor_point[self._sensors[i]]]]
        building_headers = ["Date time", "Energy consumption"]
        building_energy_consumption = pd.concat(building_measures, keys=building_headers, axis=1)

        """
        sensor = ConcordiaEnergySensor(self._sensors[i])
        sensor_exist = False
        for j in range(len(obj.sensors)):
          if obj.sensors[j].name is sensor.name:
            obj.sensors[j].add_period(building_energy_consumption)
            sensor_exist = True
            break
        if not sensor_exist:
          sensor.add_period(building_energy_consumption)
          obj.sensors.append(sensor)
        """



"""
gpandas module parses geopandas input table and import the geometry into the city model structure
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder:  Milad Aghamohamadnia --- milad.aghamohamadnia@concordia.ca
"""

import trimesh
import trimesh.exchange.load
import trimesh.geometry
import trimesh.creation
import trimesh.repair
from shapely.geometry import Point
from shapely.geometry import Polygon as ShapelyPoly
from trimesh import Scene

from hub.city_model_structure.attributes.polygon import Polygon
from hub.city_model_structure.building import Building
from hub.city_model_structure.building_demand.surface import Surface
from hub.city_model_structure.city import City

import hub.helpers.constants as cte


class GPandas:
  """
  GeoPandas class
  """

  def __init__(self, dataframe, srs_name='EPSG:26911'):
    """_summary_
    Arguments:
      dataframe {Geopandas.Dataframe} -- input geometry data in geopandas table
    Keyword Arguments:
      srs_name {str} -- coordinate system of coordinate system (default: {'EPSG:26911'})
    """

    self._srs_name = srs_name
    self._city = None
    self._scene = dataframe
    self._scene = self._scene.to_crs(self._srs_name)
    min_x, min_y, max_x, max_y = self._scene.total_bounds
    self._lower_corner = [min_x, min_y, 0]
    self._upper_corner = [max_x, max_y, 0]

  @property
  def scene(self) -> Scene:
    """
    Get GeoPandas scene
    """
    return self._scene

  @property
  def city(self) -> City:
    """
    Get city out of a GeoPandas Table
    """
    if self._city is None:
      self._city = City(self._lower_corner, self._upper_corner, self._srs_name)
      lod = 0
      for scene_index, bldg in self._scene.iterrows():
        polygon = bldg.geometry
        height = float(bldg['height'])
        building_mesh = trimesh.creation.extrude_polygon(polygon, height)
        trimesh.repair.fill_holes(building_mesh)
        trimesh.repair.fix_winding(building_mesh)
        year_of_construction = int(bldg['year_built'])
        name = bldg['name']
        lod = 1
        if year_of_construction > 2000:
          function = cte.RESIDENTIAL
        else:
          function = cte.INDUSTRY
        surfaces = []
        for face_index, face in enumerate(building_mesh.faces):
          points = []
          for vertex_index in face:
            points.append(building_mesh.vertices[vertex_index])
          solid_polygon = Polygon(points)
          perimeter_polygon = solid_polygon
          surface = Surface(solid_polygon, perimeter_polygon)
          surfaces.append(surface)
        building = Building(name, surfaces, year_of_construction, function, terrains=None)
        self._city.add_city_object(building)
      self._city.level_of_detail.geometry = lod
    return self._city

  @staticmethod
  def resize_polygon(poly, factor=0.10, expand=False) -> ShapelyPoly:
    """
    returns the shapely polygon which is smaller or bigger by passed factor.
    Arguments:
    poly {shapely.geometry.Polygon} -- an input geometry in shapely polygon format

    Keyword Arguments:
      factor {float} -- factor of expansion (default: {0.10})
      expand {bool} -- If expand = True , then it returns bigger polygon, else smaller (default: {False})

    Returns:
      {shapely.geometry.Polygon} -- output geometry in shapely polygon format
    """
    xs = list(poly.exterior.coords.xy[0])
    ys = list(poly.exterior.coords.xy[1])
    x_center = 0.5 * min(xs) + 0.5 * max(xs)
    y_center = 0.5 * min(ys) + 0.5 * max(ys)
    min_corner = Point(min(xs), min(ys))
    center = Point(x_center, y_center)
    shrink_distance = center.distance(min_corner) * factor

    if expand:
      poly_resized = poly.buffer(shrink_distance)  # expand
    else:
      poly_resized = poly.buffer(-shrink_distance)  # shrink
    return poly_resized



"""Definition of meta model 'greenerycatalog'."""
from functools import partial
import pyecore.ecore as Ecore
from pyecore.ecore import *


name = 'greenerycatalog'
nsURI = 'http://ca.concordia/greenerycatalog'
nsPrefix = 'greenery'

eClass = EPackage(name=name, nsURI=nsURI, nsPrefix=nsPrefix)

eClassifiers = {}
getEClassifier = partial(Ecore.getEClassifier, searchspace=eClassifiers)
Management = EEnum('Management', literals=['Intensive', 'Extensive', 'SemiIntensive', 'NA'])

Roughness = EEnum('Roughness', literals=['VeryRough', 'Rough',
                  'MediumRough', 'MediumSmooth', 'Smooth', 'VerySmooth'])


class Soil(EObject, metaclass=MetaEClass):

    name = EAttribute(eType=EString, unique=True, derived=False, changeable=True)
    roughness = EAttribute(eType=Roughness, unique=True, derived=False,
                           changeable=True, default_value=Roughness.MediumRough)
    conductivityOfDrySoil = EAttribute(
        eType=EString, unique=True, derived=False, changeable=True, default_value='1.0 W/(m*K)')
    densityOfDrySoil = EAttribute(eType=EString, unique=True, derived=False,
                                  changeable=True, default_value='1100 kg/m³')
    specificHeatOfDrySoil = EAttribute(
        eType=EString, unique=True, derived=False, changeable=True, default_value='1200 J/(kg*K)')
    thermalAbsorptance = EAttribute(eType=EString, unique=True,
                                    derived=False, changeable=True, default_value='0.9')
    solarAbsorptance = EAttribute(eType=EString, unique=True,
                                  derived=False, changeable=True, default_value='0.7')
    visibleAbsorptance = EAttribute(eType=EString, unique=True,
                                    derived=False, changeable=True, default_value='0.75')
    saturationVolumetricMoistureContent = EAttribute(
        eType=EString, unique=True, derived=False, changeable=True, default_value='0.0')
    residualVolumetricMoistureContent = EAttribute(
        eType=EString, unique=True, derived=False, changeable=True, default_value='0.05')
    initialVolumetricMoistureContent = EAttribute(
        eType=EString, unique=True, derived=False, changeable=True, default_value='0.1')

    def __init__(self, *, name=None, roughness=None, conductivityOfDrySoil=None, densityOfDrySoil=None, specificHeatOfDrySoil=None, thermalAbsorptance=None, solarAbsorptance=None, visibleAbsorptance=None, saturationVolumetricMoistureContent=None, residualVolumetricMoistureContent=None, initialVolumetricMoistureContent=None):
        # if kwargs:
        #    raise AttributeError('unexpected arguments: {}'.format(kwargs))

        super().__init__()

        if name is not None:
            self.name = name

        if roughness is not None:
            self.roughness = roughness

        if conductivityOfDrySoil is not None:
            self.conductivityOfDrySoil = conductivityOfDrySoil

        if densityOfDrySoil is not None:
            self.densityOfDrySoil = densityOfDrySoil

        if specificHeatOfDrySoil is not None:
            self.specificHeatOfDrySoil = specificHeatOfDrySoil

        if thermalAbsorptance is not None:
            self.thermalAbsorptance = thermalAbsorptance

        if solarAbsorptance is not None:
            self.solarAbsorptance = solarAbsorptance

        if visibleAbsorptance is not None:
            self.visibleAbsorptance = visibleAbsorptance

        if saturationVolumetricMoistureContent is not None:
            self.saturationVolumetricMoistureContent = saturationVolumetricMoistureContent

        if residualVolumetricMoistureContent is not None:
            self.residualVolumetricMoistureContent = residualVolumetricMoistureContent

        if initialVolumetricMoistureContent is not None:
            self.initialVolumetricMoistureContent = initialVolumetricMoistureContent


class Plant(EObject, metaclass=MetaEClass):

    name = EAttribute(eType=EString, unique=True, derived=False, changeable=True)
    height = EAttribute(eType=EString, unique=True, derived=False,
                        changeable=True, default_value='0.1 m')
    leafAreaIndex = EAttribute(eType=EString, unique=True, derived=False,
                               changeable=True, default_value='2.5')
    leafReflectivity = EAttribute(eType=EString, unique=True,
                                  derived=False, changeable=True, default_value='0.1')
    leafEmissivity = EAttribute(eType=EString, unique=True, derived=False,
                                changeable=True, default_value='0.9')
    minimalStomatalResistance = EAttribute(
        eType=EString, unique=True, derived=False, changeable=True, default_value='100.0  s/m')
    co2Sequestration = EAttribute(eType=EString, unique=True, derived=False,
                                  changeable=True, default_value='kgCO₂eq')
    growsOn = EReference(ordered=True, unique=True, containment=False, derived=False, upper=-1)

    def __init__(self, *, name=None, height=None, leafAreaIndex=None, leafReflectivity=None, leafEmissivity=None, minimalStomatalResistance=None, growsOn=None, co2Sequestration=None):
        # if kwargs:
        #    raise AttributeError('unexpected arguments: {}'.format(kwargs))

        super().__init__()

        if name is not None:
            self.name = name

        if height is not None:
            self.height = height

        if leafAreaIndex is not None:
            self.leafAreaIndex = leafAreaIndex

        if leafReflectivity is not None:
            self.leafReflectivity = leafReflectivity

        if leafEmissivity is not None:
            self.leafEmissivity = leafEmissivity

        if minimalStomatalResistance is not None:
            self.minimalStomatalResistance = minimalStomatalResistance

        if co2Sequestration is not None:
            self.co2Sequestration = co2Sequestration

        if growsOn:
            self.growsOn.extend(growsOn)


class SupportEnvelope(EObject, metaclass=MetaEClass):

    roughness = EAttribute(eType=Roughness, unique=True, derived=False,
                           changeable=True, default_value=Roughness.MediumRough)
    solarAbsorptance = EAttribute(eType=EDouble, unique=True,
                                  derived=False, changeable=True, default_value=0.0)
    conductivity = EAttribute(eType=EDouble, unique=True, derived=False,
                              changeable=True, default_value=0.0)
    visibleAbsorptance = EAttribute(eType=EDouble, unique=True,
                                    derived=False, changeable=True, default_value=0.0)
    specificHeat = EAttribute(eType=EDouble, unique=True, derived=False,
                              changeable=True, default_value=0.0)
    density = EAttribute(eType=EDouble, unique=True, derived=False,
                         changeable=True, default_value=0.0)
    thermalAbsorptance = EAttribute(eType=EDouble, unique=True,
                                    derived=False, changeable=True, default_value=0.0)

    def __init__(self, *, roughness=None, solarAbsorptance=None, conductivity=None, visibleAbsorptance=None, specificHeat=None, density=None, thermalAbsorptance=None):
        # if kwargs:
        #    raise AttributeError('unexpected arguments: {}'.format(kwargs))

        super().__init__()

        if roughness is not None:
            self.roughness = roughness

        if solarAbsorptance is not None:
            self.solarAbsorptance = solarAbsorptance

        if conductivity is not None:
            self.conductivity = conductivity

        if visibleAbsorptance is not None:
            self.visibleAbsorptance = visibleAbsorptance

        if specificHeat is not None:
            self.specificHeat = specificHeat

        if density is not None:
            self.density = density

        if thermalAbsorptance is not None:
            self.thermalAbsorptance = thermalAbsorptance


class GreeneryCatalog(EObject, metaclass=MetaEClass):

    name = EAttribute(eType=EString, unique=True, derived=False, changeable=True)
    description = EAttribute(eType=EString, unique=True, derived=False, changeable=True)
    source = EAttribute(eType=EString, unique=True, derived=False, changeable=True)
    plantCategories = EReference(ordered=True, unique=True,
                                 containment=True, derived=False, upper=-1)
    vegetationCategories = EReference(ordered=True, unique=True,
                                      containment=True, derived=False, upper=-1)
    soils = EReference(ordered=True, unique=True, containment=True, derived=False, upper=-1)

    def __init__(self, *, name=None, description=None, source=None, plantCategories=None, vegetationCategories=None, soils=None):
        # if kwargs:
        #    raise AttributeError('unexpected arguments: {}'.format(kwargs))

        super().__init__()

        if name is not None:
            self.name = name

        if description is not None:
            self.description = description

        if source is not None:
            self.source = source

        if plantCategories:
            self.plantCategories.extend(plantCategories)

        if vegetationCategories:
            self.vegetationCategories.extend(vegetationCategories)

        if soils:
            self.soils.extend(soils)


class PlantCategory(EObject, metaclass=MetaEClass):
    """Excluding (that is non-overlapping) categories like Trees, Hedeges, Grasses that help users finding a specific biol. plant species."""
    name = EAttribute(eType=EString, unique=True, derived=False, changeable=True)
    plants = EReference(ordered=True, unique=True, containment=True, derived=False, upper=-1)

    def __init__(self, *, name=None, plants=None):
        # if kwargs:
        #    raise AttributeError('unexpected arguments: {}'.format(kwargs))

        super().__init__()

        if name is not None:
            self.name = name

        if plants:
            self.plants.extend(plants)


class IrrigationSchedule(EObject, metaclass=MetaEClass):

    name = EAttribute(eType=EString, unique=True, derived=False, changeable=True)

    def __init__(self, *, name=None):
        # if kwargs:
        #    raise AttributeError('unexpected arguments: {}'.format(kwargs))

        super().__init__()

        if name is not None:
            self.name = name


class Vegetation(EObject, metaclass=MetaEClass):
    """Plant life or total plant cover (as of an area)"""
    name = EAttribute(eType=EString, unique=True, derived=False, changeable=True)
    thicknessOfSoil = EAttribute(eType=EString, unique=True, derived=False,
                                 changeable=True, default_value='20 cm')
    management = EAttribute(eType=Management, unique=True, derived=False,
                            changeable=True, default_value=Management.NA)
    airGap = EAttribute(eType=EString, unique=True, derived=False,
                        changeable=True, default_value='0.0 cm')
    soil = EReference(ordered=True, unique=True, containment=False, derived=False)
    plants = EReference(ordered=True, unique=True, containment=True, derived=False, upper=-1)

    def __init__(self, *, name=None, thicknessOfSoil=None, soil=None, plants=None, management=None, airGap=None):
        # if kwargs:
        #    raise AttributeError('unexpected arguments: {}'.format(kwargs))

        super().__init__()

        if name is not None:
            self.name = name

        if thicknessOfSoil is not None:
            self.thicknessOfSoil = thicknessOfSoil

        if management is not None:
            self.management = management

        if airGap is not None:
            self.airGap = airGap

        if soil is not None:
            self.soil = soil

        if plants:
            self.plants.extend(plants)


class VegetationCategory(EObject, metaclass=MetaEClass):
    """Excluding (that is non-overlapping) categories to help users finding a specific vegetation template."""
    name = EAttribute(eType=EString, unique=True, derived=False, changeable=True)
    vegetationTemplates = EReference(ordered=True, unique=True,
                                     containment=True, derived=False, upper=-1)

    def __init__(self, *, vegetationTemplates=None, name=None):
        # if kwargs:
        #    raise AttributeError('unexpected arguments: {}'.format(kwargs))

        super().__init__()

        if name is not None:
            self.name = name

        if vegetationTemplates:
            self.vegetationTemplates.extend(vegetationTemplates)


class PlantPercentage(EObject, metaclass=MetaEClass):

    percentage = EAttribute(eType=EString, unique=True, derived=False,
                            changeable=True, default_value='100')
    plant = EReference(ordered=True, unique=True, containment=False, derived=False)

    def __init__(self, *, percentage=None, plant=None):
        # if kwargs:
        #    raise AttributeError('unexpected arguments: {}'.format(kwargs))

        super().__init__()

        if percentage is not None:
            self.percentage = percentage

        if plant is not None:
            self.plant = plant



"""
Bus stop module
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""

from typing import Union
from hub.city_model_structure.transport.bus_node import BusNode
from hub.city_model_structure.transport.fast_charging_infrastructure import FastChargingInfrastructure
from hub.city_model_structure.attributes.schedule import Schedule


class BusStop(BusNode):
  """
  BusStop class
  """

  def __init__(self, name, coordinates, edges=None):
    super().__init__(name, coordinates, edges=edges, node_type='BusStop')
    self._time_table = None
    self._average_hourly_passengers_demand = None
    self._fast_charging_infrastructure = None
    self._waiting_time = None

  @property
  def time_table(self):
    """
    Add explanation here
    :return: add type of variable here
    """
    return self._time_table

  @property
  def average_hourly_passengers_demand(self) -> Schedule:
    """
    Add explanation here
    :return: Schedule
    """
    return self._average_hourly_passengers_demand

  @property
  def fast_charging_infrastructure(self) -> Union[None, FastChargingInfrastructure]:
    """
    Add explanation here
    :return: FastChargingInfrastructure
    """
    return self._fast_charging_infrastructure

  @property
  def waiting_time(self):
    """
    Add explanation here
    :return: add type of variable here
    """
    return self._waiting_time



"""
EpwWeatherParameters class to extract weather parameters from a defined region in .epw format (EnergyPlus Weather)
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""

import sys
from pathlib import Path
import pandas as pd
import hub.helpers.constants as cte


class EpwWeatherParameters:
  """
  EpwWeatherParameters class
  """
  def __init__(self, city, path, file_name):
    self._weather_values = None
    self._city = city
    self._path = Path(path / file_name)

    try:
      file = open(self._path, 'r')
      line = file.readline().split(',')
      city.climate_reference_city = line[1]
      city.latitude = line[6]
      city.longitude = line[7]
      city.time_zone = line[8]
      for i in range(0, 2):
        _ = file.readline().split(',')
      line = file.readline().split(',')
      number_records = int(line[1])
      depth_measurement_ground_temperature = []
      ground_temperature = []
      for i in range(0, number_records):
        depth_measurement_ground_temperature.append(line[i*16+2])
        temperatures = []
        for j in range(0, 12):
          temperatures.append(line[i*16+j+6])
        ground_temperature.append(temperatures)
      file.close()
    except SystemExit:
      sys.stderr.write(f'Error: weather file {self._path} not found. Please download it from '
                       f'https://energyplus.net/weather and place it in folder data\\weather\\epw\n')
      sys.exit()

    try:
      self._weather_values = pd.read_csv(self._path, header=0, skiprows=7,
                                         names=['year', 'month', 'day', 'hour', 'minute',
                                                'data_source_and_uncertainty_flags',
                                                'dry_bulb_temperature_c', 'dew_point_temperature_c',
                                                'relative_humidity_perc',
                                                'atmospheric_station_pressure_pa',
                                                'extraterrestrial_horizontal_radiation_wh_m2',
                                                'extraterrestrial_direct_normal_radiation_wh_m2',
                                                'horizontal_infrared_radiation_intensity_wh_m2',
                                                'global_horizontal_radiation_wh_m2', 'direct_normal_radiation_wh_m2',
                                                'diffuse_horizontal_radiation_wh_m2',
                                                'global_horizontal_illuminance_lux',
                                                'direct_normal_illuminance_lux', 'diffuse_horizontal_illuminance_lux',
                                                'zenith_luminance_cd_m2',
                                                'wind_direction_deg', 'wind_speed_m_s', 'total_sky_cover',
                                                'opaque_sky_cover', 'visibility_km',
                                                'ceiling_heigh_m_s', 'present_weather_observation',
                                                'present_weather_codes',
                                                'precipitable_water_mm', 'aerosol_optical_depth_10_3_ths',
                                                'snow_depth_cm',
                                                'days_since_last_snowfall', 'albedo', 'liquid_precipitation_depth_mm',
                                                'liquid_precipitation_quality_hr'])
    except SystemExit:
      sys.stderr.write(f'Error: wrong formatting of weather file {self._path}\n')
      sys.exit()

    for building in self._city.buildings:
      if cte.HOUR in building.external_temperature:
        del building.external_temperature[cte.HOUR]
      new_value = pd.DataFrame(self._weather_values[['dry_bulb_temperature_c']].to_numpy(), columns=['epw'])
      number_invalid_records = new_value[new_value.epw == 99.9].count().epw
      if number_invalid_records > 0:
        sys.stderr.write(f'Warning: {self._path} invalid records (value of 99.9) in dry bulb temperature\n')
      if cte.HOUR not in building.external_temperature:
        building.external_temperature[cte.HOUR] = new_value
      else:
        pd.concat([building.external_temperature[cte.HOUR], new_value], axis=1)

      new_value = pd.DataFrame(self._weather_values[['global_horizontal_radiation_wh_m2']].to_numpy(), columns=['epw'])
      number_invalid_records = new_value[new_value.epw == 9999].count().epw
      if number_invalid_records > 0:
        sys.stderr.write(f'Warning: {self._path} invalid records (value of 9999) in global horizontal radiation\n')
      if cte.HOUR not in building.global_horizontal:
        building.global_horizontal[cte.HOUR] = new_value
      else:
        pd.concat([building.global_horizontal[cte.HOUR], new_value], axis=1)

      new_value = pd.DataFrame(self._weather_values[['diffuse_horizontal_radiation_wh_m2']].to_numpy(), columns=['epw'])
      number_invalid_records = new_value[new_value.epw == 9999].count().epw
      if number_invalid_records > 0:
        sys.stderr.write(f'Warning: {self._path} invalid records (value of 9999) in diffuse horizontal radiation\n')
      if cte.HOUR not in building.diffuse:
        building.diffuse[cte.HOUR] = new_value
      else:
        pd.concat([building.diffuse[cte.HOUR], new_value], axis=1)

      new_value = pd.DataFrame(self._weather_values[['direct_normal_radiation_wh_m2']].to_numpy(), columns=['epw'])
      number_invalid_records = new_value[new_value.epw == 9999].count().epw
      if number_invalid_records > 0:
        sys.stderr.write(f'Warning: {self._path} invalid records (value of 9999) in direct horizontal radiation\n')
      if cte.HOUR not in building.beam:
        building.beam[cte.HOUR] = new_value
      else:
        pd.concat([building.beam[cte.HOUR], new_value], axis=1)



"""
Crossing module
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
Code contributors: Guille guille.gutierrezmorote@concordia.ca
"""

import ast
from typing import List, Union
from hub.city_model_structure.transport.traffic_node import TrafficNode


class Crossing(TrafficNode):
  """
  Crossing class
  """

  def __init__(self, name, coordinates, priority, width, shape=None, edges=None):
    super().__init__(name, coordinates, edges=edges, node_type='Crossing')
    self._priority = priority
    self._width = width
    self._shape = shape

  @property
  def priority(self) -> Union[None, bool]:
    """
    Get whether the pedestrians have priority over the vehicles
    :return: None or bool
    """
    return self._priority

  @priority.setter
  def priority(self, value):
    """
    Set whether the pedestrians have priority over the vehicles
    :param value: bool
    """
    if value is not None:
      self._priority = ast.literal_eval(value)

  @property
  def width(self) -> Union[None, float]:
    """
    Get crossing width in meters
    :return: None or float
    """
    return self._width

  @width.setter
  def width(self, value):
    """
    Set crossing width in meters
    :param value: float
    """
    if value is not None:
      self._width = float(value)

  @property
  def shape(self) -> Union[None, List[List[float]]]:
    """
    Get the list of positions
    :return: None or [[x, y, (z)]]
    """
    return self._shape

  @shape.setter
  def shape(self, value):
    """
    Set the list of positions
    :param value: [[x, y, (z)]]
    """
    if value is not None:
      self._shape = [[float(i) for i in value]]



"""
Network module
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Guille Gutierrez guillermo.gutierrezmorote@concordia.ca
"""

import uuid
from typing import List

from hub.city_model_structure.city_object import CityObject
from hub.city_model_structure.attributes.edge import Edge
from hub.city_model_structure.attributes.node import Node


class Network(CityObject):
  """
  Generic network class to be used as base for the network models
  """
  def __init__(self, name, edges=None, nodes=None):
    super().__init__(name, 0)
    if nodes is None:
      nodes = []
    if edges is None:
      edges = []
    self._id = None
    self._edges = edges
    self._nodes = nodes

  @property
  def id(self):
    """
    Get network id, a universally unique identifier randomly generated
    :return: str
    """
    if self._id is None:
      self._id = uuid.uuid4()
    return self._id

  @property
  def edges(self) -> List[Edge]:
    """
    Get network edges
    :return: [Edge]
    """
    return self._edges

  @property
  def nodes(self) -> List[Node]:
    """
    Get network nodes
    :return: [Node]
    """
    return self._nodes
